import enb.plugins


class LPAQ8Plugin(enb.plugins.PluginMake):
    name = "lpaq8"
    label = "Implementation of the LPAQ8 algorithm"
    tags = {"data compression", "image", "codec"}
    contrib_authors = ["Matt Mahoney", "Alexander Ratushnyak"]
    contrib_reference_urls = ["http://mattmahoney.net/dc/"]
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Lite archiving format to write several files into a single one.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/08"

import os
import collections


class TarliteWriter:
    """Input a series of file paths and output a single file with
    all the inputs contents, plus some meta-information to reconstruct them.
    Files are stored flatly, i.e., only names are stored,
    discarding any information about their directory structure.
    Therefore, it is not possible to store two files with the same name even if
    all input paths point to different files.
    """

    def __init__(self, initial_input_paths=None):
        self.input_paths = []
        for path in (initial_input_paths if initial_input_paths else []):
            self.add_file(path)

    def add_file(self, input_path):
        """Add a file path to the list of pending ones. Note that files are
        not read until the write() method is invoked.
        """
        assert os.path.isfile(input_path), f"Input path {input_path} does not exist."
        self.input_paths.append(input_path)

    def write(self, output_path):
        """Save the current list of input paths into output_path.
        """
        assert all(os.path.isfile(p) for p in self.input_paths), \
            "All input paths must exist at the time of writing them."

        sizes_line = ",".join(str(os.path.getsize(p)) for p in self.input_paths)
        names_line = "/".join(os.path.basename(p) for p in self.input_paths)
        with open(output_path, "wb") as output_file:
            output_file.write(f"{sizes_line}\n".encode("utf-8"))
            output_file.write(f"{names_line}\n".encode("utf-8"))
            for input_path in self.input_paths:
                with open(input_path, "rb") as input_file:
                    output_file.write(input_file.read())


class TarliteReader:
    """Extract files created by :class:`TarliteWriter`.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, tarlite_path):
        self.input_path = tarlite_path

    def extract_all(self, output_dir_path):
        """Extract all files to `output_dir_path`.
        """
        bn_count = collections.defaultdict(int)
        with open(self.input_path, "rb") as input_file:
            file_sizes = [int(e) for e in input_file.readline().decode("utf-8").split(",")]
            file_names = [e.strip() for e in input_file.readline().decode("utf-8").split("/")]
            assert len(file_sizes) == len(file_names)
            for file_size, file_name in zip(file_sizes, file_names):
                count = bn_count[file_name]
                bn_count[file_name] += 1

                output_path = os.path.join(
                    output_dir_path,
                    file_name if count == 0 else f"({count})_{file_name}")
                with open(output_path, "wb") as output_file:
                    output_file.write(input_file.read(file_size))


def tarlite_files(input_paths, output_tarlite_path):
    """Take a list of input paths and combine them into a single tarlite file.
    
    """
    writer = TarliteWriter(initial_input_paths=input_paths)
    writer.write(output_path=output_tarlite_path)


def untarlite_files(input_tarlite_path, output_dir_path):
    """Take a tarlite file and output the contents into the given directory.
    The file names are preserved.
    """
    reader = TarliteReader(tarlite_path=input_tarlite_path)
    reader.extract_all(output_dir_path=output_dir_path)



#!/usr/bin/env python3
"""Plugin to extract photometry information from a file using IRAF.
"""
__author__ = "Òscar Maireles and Miguel Hernández-Cabronero"
__since__ = "2022/11/07"

import os
import enb
import tempfile
import subprocess
import pandas as pd
import sys
from astropy.io import fits


def raw_to_fits(raw_path, fits_path):
    # Load image indexed by [x,y,z]
    img = enb.isets.load_array_bsq(raw_path)
    # Reindex image to [z,y,x] and save 
    fits.PrimaryHDU(img.swapaxes(0, 2)).writeto(fits_path)


def raw_to_photometry_df(
        raw_path,
        extension=0, fwhm=3.5, sigma=6, threshold=8.0,
        min_value=3_000_000, max_value=50_000_000,
        annulus=10.0, dannulus=10.0, aperture=4.0,
        sigma_phot=0.0):
    with tempfile.NamedTemporaryFile(dir=enb.config.options.base_tmp_dir, suffix=".csv") as csv_file, \
            tempfile.NamedTemporaryFile(dir=enb.config.options.base_tmp_dir, suffix=".fits") as fits_file:
        raw_to_fits(raw_path, fits_file.name)
        slave_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "iraf_photometry_slave.py")
        invocation = f"{sys.executable} " \
                     f"{os.path.abspath(slave_path)} " \
                     f"{os.path.abspath(fits_file.name)} {os.path.abspath(csv_file.name)}"
        status, output = subprocess.getstatusoutput(invocation)
        if status != 0:
            raise Exception(f"Status = {status} != 0.\nInput=[{invocation}].\nOutput=[{output}]")
        return pd.read_csv(csv_file.name)


class LossyPhotometryExperiment(enb.icompression.LossyCompressionExperiment):
    """Lossy compression experiment that extracts photometry-based distortion metrics.
    """

    def __init__(self, codecs, threshold,
                 dataset_paths=None,
                 csv_experiment_path=None,
                 csv_dataset_path=None,
                 dataset_info_table=None,
                 overwrite_file_properties=False,
                 task_families=None):

        super().__init__(codecs=codecs, dataset_paths=dataset_paths, csv_experiment_path=csv_experiment_path,
                         csv_dataset_path=csv_dataset_path,
                         dataset_info_table=dataset_info_table, overwrite_file_properties=overwrite_file_properties,
                         task_families=task_families)

        self.threshold = threshold

    @enb.atable.column_function([
        enb.atable.ColumnProperties("original_photometry_object_count", label="Original photometry object count",
                                    plot_min=0),
        enb.atable.ColumnProperties("reconstructed_photometry_object_count",
                                    label="Reconstructed photometry object count", plot_min=0),
        enb.atable.ColumnProperties("recovered_objects", label="Recovered objects", plot_min=0),
        enb.atable.ColumnProperties("mean_magnitude_difference", label="Mean magnitude difference", plot_min=0),
        enb.atable.ColumnProperties("maximum_magnitude_difference", label="Maximum magnitude difference", plot_min=0),
        enb.atable.ColumnProperties("F1_score", label="F1 score", plot_mitrue_positive=0),
    ])
    def set_photometry_columns(self, index, row):
        original_raw_path, codec = self.index_to_path_task(index)
        reconstructed_raw_path = self.codec_results.decompression_results.reconstructed_path

        original_photometry_df = raw_to_photometry_df(raw_path=original_raw_path)
        reconstructed_photometry_df = raw_to_photometry_df(raw_path=reconstructed_raw_path)

        x_position_original = original_photometry_df.loc[:, "x"]
        x_position_reconstructed = reconstructed_photometry_df.loc[:, "x"]
        y_position_original = original_photometry_df.loc[:, "y"]
        y_position_reconstructed = reconstructed_photometry_df.loc[:, "y"]
        magnitude_original = original_photometry_df.loc[:, "magnitude"]
        magnitude_reconstructed = reconstructed_photometry_df.loc[:, "magnitude"]

        true_positive = 0
        magnitude_difference = []

        for i in range(len(x_position_reconstructed)):
            for j in range(len(x_position_original)):
                if len(x_position_original) == 1:
                    pass
                elif abs(x_position_original[j] - x_position_reconstructed[i]) < self.threshold \
                        and abs(y_position_original[j] - y_position_reconstructed[i]) < self.threshold:
                    magnitude_difference.append(abs(magnitude_original[j] - magnitude_reconstructed[i]))
                    true_positive = true_positive + 1

        false_negative = len(x_position_original) - true_positive
        false_positive = len(x_position_reconstructed) - true_positive

        row["original_photometry_object_count"] = len(original_photometry_df)
        row["reconstructed_photometry_object_count"] = len(reconstructed_photometry_df)
        row["recovered_objects"] = true_positive
        row["mean_magnitude_difference"] = sum(magnitude_difference) / len(magnitude_difference) \
            if len(magnitude_difference) else 0
        row["maximum_magnitude_difference"] = max(magnitude_difference) \
            if len(magnitude_difference) else 0
        row["F1_score"] = 2 * true_positive / (2 * true_positive + false_positive + false_negative) \
            if true_positive + false_positive + false_negative > 0 else 0

        if not len(magnitude_difference):
            enb.logger.warn(f"Warning: found zero magnitude differences for "
                            f"{original_raw_path} and {codec.label} ({codec.name})")



#!/usr/bin/env python3
"""Utils to plot data (thinking about pyplot).
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2019/09/10"

import os
import itertools
import glob
import collections
import matplotlib
from matplotlib import pyplot as plt
import numpy as np
from scipy.stats import norm
import enb



class PlottableData:
    """Base class for plottable data elements. Subclasses are instantiated
    by the analyzers to define sequences of plotting actions (with pyplot)
    that result in the desired figure.
    """
    # pylint: disable=too-many-instance-attributes

    alpha = 0.75
    legend_column_count = 1
    color = None
    marker = None
    # If "title", it is shown outside the plot, above it and centered.
    # Otherwise, it must be a matplotlib-recognized string
    legend_position = "title"

    def __init__(self, data=None, axis_labels=None, label=None,
                 extra_kwargs=None, alpha=None, legend_column_count=None,
                 legend_position=None,
                 marker=None, color=None,
                 marker_size=None):
        # pylint: disable=too-many-arguments
        self.data = data
        self.axis_labels = axis_labels
        self.label = label
        self.extra_kwargs = extra_kwargs if extra_kwargs is not None else {}
        self.alpha = alpha if alpha is not None else self.alpha
        self.legend_column_count = legend_column_count \
            if legend_column_count is not None else self.legend_column_count
        self.legend_position = legend_position \
            if legend_position is not None else self.legend_position
        self.marker = marker
        self.marker_size = marker_size
        self.color = color if color is not None else self.color

    def render(self, axes=None):
        """Render data in current figure.

        :param axes: if not None, those axes are used for plotting instead of
          plt
        """
        raise NotImplementedError

    def render_axis_labels(self, axes=None):
        """Add axis labels in current figure - don't show or write the result

        :param axes: if not None, those axes are used for plotting instead of
          plt
        """
        raise NotImplementedError

    def render_legend(self, axes=None):
        """Tell numpy to render the legend based on self.legend_position.
        In addition to numpy's legend positions, 'title' is available for
        top center outside the grid.
        """
        axes = plt if axes is None else axes
        if self.legend_position == "title":
            legend = axes.legend(loc="lower center", bbox_to_anchor=(0.5, 1),
                                 ncol=self.legend_column_count,
                                 edgecolor=(0, 0, 0, 0.2))
            facecolor = (1, 1, 1, 0)
        else:
            legend = axes.legend(
                loc=self.legend_position
                if self.legend_position is not None else "best",
                ncol=self.legend_column_count, edgecolor=(0, 0, 0, 0.2))
            facecolor = (1, 1, 1, 1)

        legend.get_frame().set_alpha(None)
        legend.get_frame().set_facecolor(facecolor)

    def __repr__(self):
        return f"{self.__class__.__name__}(color={repr(self.color)})"


class PlottableData2D(PlottableData):
    """Base class for 2D plottable data.
    """

    # pylint: disable=abstract-method

    def __init__(self, x_values, y_values,
                 x_label=None, y_label=None,
                 label=None, extra_kwargs=None,
                 remove_duplicates=False,
                 alpha=None, legend_column_count=None,
                 marker=None, color=None,
                 marker_size=None):
        """
        :param x_values, y_values: values to be plotted (only a reference is
          kept)
        :param x_label, y_label: axis labels
        :param label: line legend label
        :param extra_kwargs: extra arguments to be passed to plt.plot
        """
        # pylint: disable=too-many-arguments,too-many-locals
        assert len(x_values) == len(y_values), \
            f"Invalid lengths of x_values and y_values ({len(x_values), len(y_values)}"
        if remove_duplicates:
            found_pairs = collections.OrderedDict()
            for x_val, y_val in zip(x_values, y_values):
                found_pairs[(x_val, y_val)] = (x_val, y_val)
            if found_pairs:
                x_values, y_values = zip(*found_pairs.values())
            else:
                x_values = []
                y_values = []

        super().__init__(data=(x_values, y_values),
                         axis_labels=(x_label, y_label),
                         label=label, extra_kwargs=extra_kwargs, alpha=alpha,
                         legend_column_count=legend_column_count,
                         marker_size=marker_size, color=color,
                         marker=marker)
        self.x_values = x_values
        self.y_values = y_values
        self.x_label = x_label
        self.y_label = y_label

    def render_axis_labels(self, axes=None):
        """Show the labels in label list (if not None) or in
        self.axis_label_list (if label_list None) in the current figure.

        :param axes: if not None, those axes are used for plotting instead of
          plt
        """
        axes = plt if axes is None else axes
        if self.x_label is not None:
            axes.set_xlabel(self.x_label)
        if self.y_label is not None:
            axes.set_ylabel(self.y_label)

    def diff(self, other, ylabel_suffix="_difference"):
        """Calculate the difference with another PlattableData2D instance.
        The x values are maintained and other's y values are subtracted
        from self's.
        """
        assert len(self.x_values) == len(other.x_values)
        assert len(self.y_values) == len(other.y_values)
        assert all(s == o for s, o in zip(self.x_values, other.x_values))
        return PlottableData2D(
            x_values=self.x_values,
            y_values=[s - o for s, o in zip(self.y_values, other.y_values)],
            x_label=self.x_label,
            y_label=f"{self.y_label}{ylabel_suffix}",
            legend_column_count=self.legend_column_count)

    def shift_x(self, constant):
        """Add a constant to all x values.
        """
        self.y_values = [y + constant for y in self.y_values]

    def shift_y(self, constant):
        """Add a constant to all y values.
        """
        self.y_values = [y + constant for y in self.y_values]


class LineData(PlottableData2D):
    """Straight lines linking the defined x,y pairs.
    """

    def __init__(self, marker="o", marker_size=5, line_width=1.5, **kwargs):
        super().__init__(marker=marker, marker_size=marker_size, **kwargs)
        self.line_width = line_width

    def render(self, axes=None):
        try:
            original_kwargs = self.extra_kwargs
            extra_kwargs = dict(self.extra_kwargs)

            try:
                marker_size = extra_kwargs["ms"]
                del extra_kwargs["ms"]
            except KeyError:
                marker_size = self.marker_size

            axes.plot(self.x_values, self.y_values, label=self.label,
                      alpha=self.alpha,
                      marker=self.marker, ms=marker_size,
                      linewidth=self.line_width,
                      **extra_kwargs)
            axes = plt if axes is None else axes
            self.render_axis_labels(axes=axes)
            if self.label is not None and self.legend_column_count != 0:
                self.render_legend(axes=axes)
        finally:
            self.extra_kwargs = original_kwargs


class ScatterData(PlottableData2D):
    """Individual markers at the specified x,y positions.
    """

    def __init__(self, marker="o", alpha=0.5, marker_size=3, **kwargs):
        super().__init__(marker=marker, alpha=alpha, marker_size=marker_size,
                         **kwargs)

    def render(self, axes=None):
        axes = plt if axes is None else axes
        self.extra_kwargs["s"] = self.marker_size

        axes.scatter(self.x_values, self.y_values, label=self.label,
                     alpha=self.alpha,
                     marker=self.marker,
                     **self.extra_kwargs)
        self.render_axis_labels(axes=axes)
        if self.label is not None and self.legend_column_count != 0:
            self.render_legend(axes=axes)


class BarData(PlottableData2D):
    """Vertical (horizonal) bars at x (y) positions of height (width) y (x).
    """
    pattern = None
    vertical = True

    def __init__(self, pattern=None, vertical=True, **kwargs):
        super().__init__(**kwargs)
        self.pattern = pattern
        self.vertical = vertical

    def render(self, axes=None):
        axes = plt if axes is None else axes
        plot_fun = axes.bar if self.vertical is True else axes.barh
        plot_fun(self.x_values, self.y_values, label=self.label,
                 alpha=self.alpha if not self.pattern else self.alpha * 0.75,
                 hatch=self.pattern,
                 edgecolor=matplotlib.colors.colorConverter.to_rgba(
                     self.color, 0.9) if self.color else None,
                 **self.extra_kwargs)
        self.render_axis_labels(axes=axes)
        if self.label is not None and self.legend_column_count != 0:
            self.render_legend(axes=axes)

    def shift_y(self, constant):
        coordinate = "bottom" if self.vertical else "left"
        try:
            self.extra_kwargs[coordinate] += constant
        except KeyError:
            self.extra_kwargs[coordinate] = constant
        except AttributeError:
            self.extra_kwargs = {coordinate: constant}


class StepData(PlottableData2D):
    """Horizontal segments at the defined x,y positions.
    """
    where = "post"

    def render(self, axes=None):
        axes = plt if axes is None else axes
        axes.step(self.x_values, self.y_values, label=self.label,
                  alpha=self.alpha,
                  where=self.where, **self.extra_kwargs)
        self.render_axis_labels(axes=axes)
        if self.label is not None and self.legend_column_count != 0:
            self.render_legend(axes=axes)


class ErrorLines(PlottableData2D):
    """One or more error lines
    """

    def __init__(self, x_values, y_values, err_neg_values, err_pos_values,
                 vertical=False,
                 alpha=0.5, color=None,
                 marker_size=1, cap_size=2,
                 line_width=1, **kwargs):
        """
        :param x_values, y_values: centers of the error lines
        :param err_neg_values: list of lengths for the negative part of the error
        :param err_pos_values: list of lengths for the positive part of the error
        :param vertical: determines whether the error bars are vertical or horizontal
        """
        # pylint: disable=too-many-arguments
        super().__init__(x_values=x_values, y_values=y_values,
                         remove_duplicates=False,
                         alpha=alpha, marker_size=marker_size, **kwargs)
        self.err_neg = [v if not np.isnan(v) else 0 for v in err_neg_values]
        self.err_pos = [v if not np.isnan(v) else 0 for v in err_pos_values]
        self.cap_size = cap_size if cap_size is not None else self.cap_size
        self.line_width = line_width if line_width is not None else self.line_width
        self.vertical = vertical
        self.color = color
        self.alpha = alpha
        assert len(self.x_values) == len(self.y_values)
        assert len(self.x_values) == len(self.err_pos), (
            len(self.x_values), len(self.err_pos), self.x_values, self.err_pos)

    def render(self, axes=None):
        axes = plt if axes is None else axes
        err_argument = np.concatenate((
            np.array(self.err_neg).reshape(1, len(self.err_neg)),
            np.array(self.err_pos).reshape(1, len(self.err_pos))),
            axis=0)
        # When std is computed for constant data, nan is returned. Replace
        # those with 0 to avoid warnings.
        err_argument = err_argument[:, :len(self.x_values)]

        try:
            if self.vertical:
                axes.errorbar(self.x_values, self.y_values, yerr=err_argument,
                              fmt="-o", capsize=self.cap_size, capthick=0.5,
                              lw=0, elinewidth=self.line_width,
                              ms=self.marker_size,
                              alpha=self.alpha,
                              **self.extra_kwargs)
            else:
                axes.errorbar(self.x_values, self.y_values, xerr=err_argument,
                              fmt="-o", capsize=self.cap_size, capthick=0.5,
                              lw=0, elinewidth=self.line_width,
                              ms=self.marker_size,
                              alpha=self.alpha,
                              **self.extra_kwargs)
        except UserWarning as ex:
            raise ValueError(
                f"Error rendering {self}. x_values={repr(self.x_values)}, "
                f"y_values={repr(self.y_values)}, "
                f"self.extra_kwargs={self.extra_kwargs} "
                f"self.err_neg={self.err_neg}, "
                f"self.err_pos={self.err_pos},") from ex

        assert len(self.x_values) == len(self.y_values)


class Rectangle(PlottableData2D):
    """Render a rectangle (line only) in a given position.
    """
    alpha = 0.5

    def __init__(self, x_values, y_values, width, height, angle_degrees=0,
                 fill=False,
                 line_width=1, **kwargs):
        """
        :param x_values: a list with a single element with the x position of
          the rectangle's center
        :param y_values: a list with a single element with the y position of
          the rectangle's center
        :param width: width of the rectangle
        :param height: height of the rectangle
        """
        # pylint: disable=too-many-arguments
        super().__init__(x_values=x_values, y_values=y_values, **kwargs)
        assert width >= 0, width
        assert height >= 0, height
        assert line_width >= 0, line_width
        self.width = width
        self.height = height
        self.angle_degrees = angle_degrees
        self.fill = fill
        self.line_width = line_width

    def render(self, axes=None):
        assert len(self.x_values) == 1, self.x_values
        assert len(self.y_values) == 1, self.y_values
        axes = plt if axes is None else axes
        axes.add_patch(matplotlib.patches.Rectangle((
            self.x_values[0] - self.width / 2,
            self.y_values[0] - self.height / 2),
            width=self.width, height=self.height,
            angle=self.angle_degrees, fill=self.fill,
            color=self.color, alpha=self.alpha,
            linewidth=self.line_width))


class LineSegment(PlottableData2D):
    """Render a horizontal or vertical line segment centered at a given
    position.
    """
    alpha = 0.5

    def __init__(self, x_values, y_values, length, vertical=True, line_width=1,
                 **kwargs):
        """
        :param x_values: a list with a single element with the x position of
          the rectangle's center
        :param y_values: a list with a single element with the y position of
          the rectangle's center
        :param length: length of the line
        """
        # pylint: disable=too-many-arguments
        super().__init__(x_values=x_values, y_values=y_values, **kwargs)
        assert length >= 0, length
        assert line_width >= 0, line_width
        self.length = length
        self.line_width = line_width
        self.vertical = vertical

    @property
    def center(self):
        """Line center (the line crosses this point).
        """
        return self.x_values[0], self.y_values[0]

    def render(self, axes=None):
        assert len(self.x_values) == 1, self.x_values
        assert len(self.y_values) == 1, self.y_values
        axes = plt if axes is None else axes
        center = self.center
        if self.vertical:
            x_values = [center[0], center[0]]
            y_values = [center[1] - self.length / 2,
                        center[1] + self.length / 2]
        else:
            x_values = [center[0] - self.length / 2,
                        center[0] + self.length / 2]
            y_values = [center[1], center[1]]

        axes.plot(x_values, y_values, color=self.color, alpha=self.alpha,
                  linewidth=self.line_width)


class HorizontalBand(PlottableData2D):
    """Colored band surrounding a line function given by the provided x,
    y positions, with positive (up) and negative (down) widths.
    """
    alpha = 0.5
    degradation_band_count = 25
    show_bounding_lines = False

    def __init__(self, x_values, y_values, pos_height_values, neg_height_values,
                 show_bounding_lines=None,
                 degradation_band_count=None,
                 std_band_add_xmargin=False,
                 **kwargs):
        # pylint: disable=too-many-arguments
        super().__init__(x_values=x_values, y_values=y_values,
                         extra_kwargs=kwargs)
        self.extra_kwargs["lw"] = 0
        self.pos_height_values = np.array(pos_height_values)
        self.neg_height_values = np.array(neg_height_values)
        self.show_bounding_lines = show_bounding_lines \
            if show_bounding_lines is not None else self.show_bounding_lines
        self.degradation_band_count = degradation_band_count \
            if degradation_band_count is not None \
            else self.degradation_band_count
        self.std_band_add_xmargin = std_band_add_xmargin

    def render(self, axes=None):
        # pylint: disable=too-many-locals
        for i in range(self.degradation_band_count):
            band_fraction = i / self.degradation_band_count
            next_band_fraction = (i + 1) / self.degradation_band_count
            band_probability = \
                1 - (norm.cdf(band_fraction) - norm.cdf(-band_fraction))
            band_x_values = np.array(self.x_values)
            band_y_values = np.array(self.y_values)
            pos_height_values = np.array(self.pos_height_values)
            neg_height_values = np.array(self.neg_height_values)
            if self.std_band_add_xmargin:
                new_x_values = []
                new_y_values = []
                new_pos_heights = []
                new_neg_heights = []
                for x_value, y_value, pos_height, neg_height in zip(
                        band_x_values, band_y_values,
                        pos_height_values, neg_height_values):
                    new_x_values.extend([x_value - 0.5, x_value + 0.5])
                    new_y_values.extend([y_value, y_value])
                    new_pos_heights.extend([pos_height, pos_height])
                    new_neg_heights.extend([neg_height, neg_height])
                band_x_values = np.array(new_x_values)
                band_y_values = np.array(new_y_values)
                pos_height_values = np.array(new_pos_heights)
                neg_height_values = np.array(new_neg_heights)

            # Fill top
            axes.fill_between(
                band_x_values,
                band_y_values + pos_height_values * band_fraction,
                band_y_values + pos_height_values * next_band_fraction,
                alpha=self.alpha * band_probability,
                color=self.color, edgecolor=None, facecolor=self.color, lw=0)
            # Fill bottom
            axes.fill_between(
                band_x_values,
                band_y_values - neg_height_values * next_band_fraction,
                band_y_values - neg_height_values * band_fraction,
                alpha=self.alpha * band_probability,
                color=self.color, edgecolor=None, facecolor=self.color, lw=0)

        if self.show_bounding_lines:
            axes.plot(self.x_values, self.y_values + pos_height_values,
                      linewidth=0.5, alpha=0.68 * self.alpha, color=self.color)
            axes.plot(self.x_values, self.y_values - neg_height_values,
                      linewidth=0.5, alpha=0.68 * self.alpha, color=self.color)

        axes = plt if axes is None else axes
        self.render_axis_labels(axes=axes)
        if self.label is not None and self.legend_column_count != 0:
            plt.legend(loc="lower center", bbox_to_anchor=(0.5, 1),
                       ncol=self.legend_column_count, facecolor=(1, 1, 1, 0))


class HorizontalLine(PlottableData2D):
    """Draw a horizontal line across the whole subplot.
    """

    def __init__(self, y_position, line_width=1, line_style='-', **kwargs):
        super().__init__(**kwargs, x_values=[0], y_values=[y_position])
        self.y_position = y_position
        self.line_width = line_width
        self.line_style = line_style

    def render(self, axes=None):
        axes = plt if axes is None else axes
        axes.axhline(y=self.y_position, color=self.color,
                     linestyle=self.line_style,
                     lw=self.line_width, alpha=self.alpha)


class VerticalLine(PlottableData2D):
    """Draw a horizontal line across the whole subplot.
    """

    def __init__(self, x_position, line_width=1, line_style='-', **kwargs):
        super().__init__(**kwargs, x_values=[x_position], y_values=[0])
        self.x_position = x_position
        self.line_width = line_width
        self.line_style = line_style

    def render(self, axes=None):
        axes = plt if axes is None else axes
        axes.axvline(x=self.x_position, color=self.color,
                     linestyle=self.line_style,
                     lw=self.line_width, alpha=self.alpha)


class Histogram2D(PlottableData2D):
    """Represent the result of a 2D histogram.
    See https://matplotlib.org/stable/gallery/color/colormap_reference.html
    """
    # pylint: disable=too-many-instance-attributes

    color_map = "Reds"
    interpolation = "none"
    origin = "lower"
    alpha = 1
    aspect = "equal"
    vmin = None
    vmax = None
    show_cmap_bar = True
    no_data_color = (1, 1, 1, 0)
    bad_data_color = "magenta"

    def __init__(self, x_edges, y_edges, matrix_values, color_map=None,
                 colormap_label=None, vmin=None, vmax=None,
                 no_data_color=(1, 1, 1, 0), bad_data_color="magenta",
                 **kwargs):
        """
        :param x_edges: the edges of the histogram along the x axis.
        :param y_edges: the edges of the histogram along the y axis.
        :param matrix_values: values of the histogram (2d array of dimensions
          given by the length of x_edges and y_edges).
        :param no_data_color: color shown when no counts are found in a bin
        :param bad_data_color: color shown when nan is found in a bin
        :param vmin: minimum value considered in the histogram
        :param vmax: minimum value considered in the histogram
        :param kwargs: additional parameters passed to the parent class
          initializer
        """
        # pylint: disable=too-many-arguments
        super().__init__(x_values=tuple(x - 0.5 for x in range(len(x_edges))),
                         y_values=tuple(y - 0.5 for y in range(len(y_edges))),
                         **kwargs)
        self.x_edges = x_edges
        self.y_edges = y_edges
        self.matrix_values = matrix_values
        self.colormap_label = colormap_label
        self.vmin = vmin
        self.vmax = vmax
        self.color_map = color_map if color_map is not None else self.color_map
        self.bad_data_color = bad_data_color
        self.no_data_color = no_data_color

    def render(self, axes=None):
        axes = plt if axes is None else axes

        cmap = matplotlib.cm.get_cmap(self.color_map).copy()
        cmap.set_under(color=self.no_data_color)
        cmap.set_over(color=self.bad_data_color)
        cmap.set_bad(color=self.bad_data_color)

        axes_image = axes.imshow(self.matrix_values.swapaxes(0,1), cmap=cmap,
                                 origin=self.origin,
                                 interpolation=self.interpolation,
                                 alpha=self.alpha, aspect=self.aspect,
                                 vmin=self.vmin, vmax=self.vmax)
        if self.show_cmap_bar:
            plt.colorbar(axes_image, label=self.colormap_label,
                         ticks=list(np.linspace(self.vmin, self.vmax, 5))
                         if self.vmin is not None and self.vmax is not None else None,
                         fraction=0.1, ax=axes)
        if self.y_label:
            axes.set_ylabel(self.y_label)

class Table(PlottableData2D):
    """Display a 2D table of data.
    """
    def __init__(self, x_values, y_values, cell_text, x_label=None, y_label=None,
                 cell_alignment="center",
                 col_header_aligment="center",
                 row_header_alignment="left",
                 edges="open",
                 highlight_best_row=None,
                 highlight_best_column=None):
        """
        :param x_values: list of column headers
        :param y_values: list of row headers
        :param cell_text: list of lists containing the cell text to display
        :param cell_alignment: text alignment for the data cells
        :param col_header_alignment: text alignment for the column headers
        :param row_header_alignment: text alignment for the row headers
        :param edges: argument passed to plt.table
          (substring of 'BRTL' or {'open', 'closed', 'horizontal', 'vertical'})
        :param highlight_best_row: if not None, it must be either "low" or "high". In those cases, the
          best (lowest or highest) value of each column is highlighted.
        :param highlight_best_column: if not None, it must be either "low" or "high". In those cases, the
          best (lowest or highest) value of each row is highlighted.
        """
        super().__init__(x_values=[None], y_values=[None], x_label=x_label, y_label=y_label)
        self.x_values = x_values
        self.y_values = y_values
        self.cell_text = cell_text
        self.cell_alignment = cell_alignment
        self.col_header_alignment = col_header_aligment
        self.row_header_alignment = row_header_alignment
        self.edges = edges
        self.highlight_best_row = highlight_best_row
        self.highlight_best_column = highlight_best_column

    def render(self, axes=None):
        axes = plt if axes is None else axes
        axes.axis('tight')
        axes.axis('off')

        table = axes.table(cellText=self.cell_text,colLabels=self.x_values,rowLabels=self.y_values,
                   loc="center", cellLoc=self.cell_alignment,
                   rowLoc=self.row_header_alignment, colLoc=self.col_header_alignment,
                   bbox=[0,0,1,1], edges=self.edges)

        # Make headers bold and the background transparent
        for (row, col), cell in table.get_celld().items():
            if (row == 0) or (col == -1):
                cell.set_text_props(
                    fontproperties=matplotlib.font_manager.FontProperties(weight='bold'))
            cell.set_facecolor((1,1,0,0))

        # Highlight the best column in each row
        if self.highlight_best_column is not None:
            for row_index, row in enumerate(self.cell_text):
                row_values = [float(c) for c in row]
                if self.highlight_best_column.lower() == "low":
                    best_value = min(row_values)
                else:
                    assert self.highlight_best_column.lower() == "high", f"Invalid {self.highlight_best_column=}"
                    best_value = max(row_values)
                for col_index, col_value in enumerate(row_values):
                    if col_value == best_value:
                        table.get_celld()[(row_index+1, col_index)].set_text_props(
                            fontproperties=matplotlib.font_manager.FontProperties(weight='bold'))

        # Highlight the best row in each column
        if self.highlight_best_row is not None:
            column_count = len(self.cell_text[0])
            for col_index in range(column_count):
                col_values = [float(r[col_index]) for r in self.cell_text]
                if self.highlight_best_row == "low":
                    best_value = min(col_values)
                else:
                    assert self.highlight_best_row == "high", f"Invalid {self.highlight_best_row=}"
                    best_value = max(col_values)
                for row_index, row in enumerate(self.cell_text):
                    if float(row[col_index]) == best_value:
                        table.get_celld()[(row_index+1, col_index)].set_text_props(
                            fontproperties=matplotlib.font_manager.FontProperties(weight='bold'))


def get_available_styles():
    """Get a list of all styles available for plotting. It includes installed
    matplotlib styles plus custom styles
    """
    return sorted(
        itertools.chain(get_matlab_styles(), get_local_styles(), ["default"]))


def get_matlab_styles():
    """Return the list of installed matlab styles.
    """
    return sorted(
        [str(s) for s in matplotlib.style.available if s[0] != "_"] + ["xkcd"])


def get_local_styles():
    """Get the list of basenames of all styles in the enb/config/mpl_styles
    folder.
    """
    return sorted(os.path.basename(p)
                  for p in glob.glob(
        os.path.join(enb.enb_installation_dir, "config", "mpl_styles", "*"))
                  if os.path.isfile(p) and os.path.basename(p) != "README.md")



#!/usr/bin/env python3
"""Unit tests for sets.py
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2019/10/05"

import unittest
import tempfile
import glob
import os
import shutil
import filecmp
import numpy as np

import enb
import enb.atable
import enb.sets as sets
from enb.config import options


class TestSets(unittest.TestCase):

    def test_file_properties(self):
        """Test that file properties are correctly obtained and retrieved.
        """
        target_indices = [p for p in glob.glob(os.path.join(enb.calling_script_dir, "*.py"))
                          if os.path.isfile(p)]

        # dataset_df = get_result_df()
        with tempfile.NamedTemporaryFile(suffix=".csv") as tmp_file:
            dataset_properties_table = sets.FilePropertiesTable(
                csv_support_path=tmp_file.name)

            # Attempt loading from an empty file, verify it is empty because fill=False
            try:
                empty_property_table = dataset_properties_table.get_df(
                    fill=False, target_indices=target_indices)
            except ValueError:
                assert len(target_indices) == 0

            assert len(empty_property_table) == 0, empty_property_table
            assert empty_property_table.isnull().all().all()

            # Run the actual loading sequence
            dataset_properties_df = dataset_properties_table.get_df(
                target_indices=target_indices)
            assert len(dataset_properties_df) == len(target_indices)

            # Obtain again, forcing load from the temporary file without any additional computations
            new_df = dataset_properties_table.get_df(
                target_indices=target_indices, fill=False, overwrite=False)
            assert (dataset_properties_df.columns == new_df.columns).all()

            for c in dataset_properties_df.columns:
                try:
                    if not (dataset_properties_df[c] == new_df[c]).all():
                        # Floating point values might be unstable
                        try:
                            assert np.abs(dataset_properties_df[c] - new_df[c]).max() < 1e-12, (
                            dataset_properties_df[c], new_df[c])
                        except TypeError:
                            # Stability within dictionaries is not verified,
                            # but only dictionaries can raise this error
                            assert (dataset_properties_df[c].apply(
                                lambda c: isinstance(c, dict))).all()
                except ValueError as ex:
                    raise RuntimeError(
                        "The original and loaded datasets differ") from ex

    def test_trivial_version_table(self):
        """Test versioning with a table that simply copies the original file.
        """

        class TrivialVersionTable(sets.FileVersionTable):
            """Trivial FileVersionTable that makes an identical copy of the original
            """

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.dataset_files_extension = "py"

            def version(self, input_path, output_path, row):
                os.makedirs(os.path.dirname(output_path), exist_ok=True)
                shutil.copy(input_path, output_path)
                assert os.path.getsize(input_path) == os.path.getsize(
                    output_path)

        with tempfile.TemporaryDirectory() as tmp_dir:
            options.persistence_dir = tmp_dir
            fpt = sets.FilePropertiesTable(base_dir=options.project_root)
            fpt.dataset_files_extension = "py"
            fpt_df = fpt.get_df()
            fpt_df["original_file_path"] = fpt_df["file_path"]
            tvt = TrivialVersionTable(version_base_dir=tmp_dir,
                                      version_name="trivial",
                                      original_base_dir=options.project_root)
            tvt_df = tvt.get_df()
            for input_path in (
                    p for p in glob.glob(
                os.path.join(options.project_root, "**", "*.py"), recursive=True)
                    if os.path.isfile(p)):
                assert filecmp.cmp(input_path,
                                   tvt.original_to_versioned_path(input_path))

            lsuffix = "_original"
            rsuffix = f"_{tvt.version_name}"
            joint_df = fpt_df.set_index("original_file_path").join(
                tvt_df.set_index("original_file_path"),
                lsuffix=lsuffix, rsuffix=rsuffix)

            assert not np.any(joint_df.map(lambda x: x is None))

            for column in [c for c in joint_df.columns.values if lsuffix in c]:
                if column.replace(lsuffix, "") in fpt.indices:
                    continue
                version_column = column.replace(lsuffix, rsuffix)
                if not column.startswith("corpus"):
                    assert np.all(joint_df[column] == joint_df[version_column]) \
                           or column.startswith("row_created") \
                           or column.startswith("row_update"), \
                        f"Columns {column} and {version_column} differ: " \
                        f"{joint_df[joint_df[column] != joint_df[version_column]][[column, version_column]].iloc[0]}"


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Slave script to run iraf as a separate process, so that it is automatically cleaned up 
and memory is not hogged.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2022/11/16"

import os
import tempfile
import contextlib
import argparse

default_iraf_folder = "/usr/local/bin/cl"
custom_iraf_folder = "/usr/lib/iraf"
iraf_folder = default_iraf_folder

if "iraf" not in os.environ or not os.path.isdir(os.environ["iraf"]):
    if not os.path.isdir(default_iraf_folder):
        if not os.path.isdir(custom_iraf_folder):
            raise RuntimeError(f"Cannot find iraf in either the default ({default_iraf_folder}) "
                               f"nor the custom folders ({custom_iraf_folder}). "
                               f"Please modify custom_iraf_folder in {os.path.abspath(__file__)} "
                               f"to match your installation")
        iraf_folder = custom_iraf_folder
    os.environ["iraf"] = iraf_folder

from pyraf import iraf


def fits_to_csv(
        fits_path, csv_path,
        extension=0, fwhm=3.5, sigma=6, threshold=8.0,
        min_value=3_000_000, max_value=50_000_000,
        annulus=10.0, dannulus=10.0, aperture=4.0,
        sigma_phot=0.0):
    """Apply IRAF to the raw file in `raw_path` and store a CSV file into `csv_path`.
    """
    if os.path.exists(csv_path):
        os.remove(csv_path)

    with tempfile.NamedTemporaryFile() as daofind_file, \
            tempfile.NamedTemporaryFile() as phot_file, \
            tempfile.NamedTemporaryFile() as daofind_params_file, \
            tempfile.NamedTemporaryFile() as phot_params_file, \
            tempfile.NamedTemporaryFile() as filtered_phot_file:
        for tmp_file in (daofind_file, phot_file, daofind_params_file,
                         phot_params_file, filtered_phot_file):
            os.remove(tmp_file.name)

        # Find objects in the FITS file
        iraf.noao.digiphot(_doprint=0)
        iraf.noao.digiphot.phot(_doprint=0)
        iraf.noao.digiphot.daophot(_doprint=0)
        iraf.digiphot.ptools(_doprint=0)
        iraf.datapars.datamin = min_value
        iraf.datapars.datamax = max_value
        iraf.findpars.threshold = threshold
        iraf.findpars.nsigma = sigma
        iraf.datapars.fwhmpsf = fwhm
        iraf.datapars.sigma = sigma
        iraf.daofind.setParam('fwhmpsf', fwhm)
        iraf.daofind.setParam('output', daofind_file.name)
        iraf.daofind.setParam('sigma', sigma)
        iraf.daofind.setParam('thresh', threshold)
        iraf.daofind.setParam('verify', "no")
        iraf.daofind.setParam('cache', "no")
        iraf.daofind.setParam('datamin', min_value)
        iraf.daofind.setParam('datamax', max_value)
        iraf.daofind.saveParList(filename=daofind_params_file.name)
        with open(os.devnull, "w") as devnull, contextlib.redirect_stdout(devnull):
            iraf.daofind(f"{fits_path}[{extension}]", ParList=daofind_params_file.name)

        # Obtain photometry results
        iraf.fitskypars.salgorithm = "mode"
        iraf.fitskypars.dannulus = dannulus
        iraf.photpars.weighting = "constant"
        iraf.phot.interactive = "no"
        iraf.phot.verify = "no"
        iraf.phot.update = "no"
        iraf.phot.setParam('calgorithm', "none")
        iraf.phot.setParam('salgorithm', "mode")
        iraf.phot.setParam('annulus', annulus)
        iraf.phot.setParam('dannulus', dannulus)
        iraf.phot.setParam('apertures', aperture)
        iraf.phot.setParam('sigma', sigma_phot)
        iraf.phot.setParam('datamin', "INDEF")
        iraf.phot.setParam('datamax', "INDEF")
        iraf.phot.saveParList(filename=phot_params_file.name)
        with open(os.devnull, "w") as devnull, contextlib.redirect_stdout(devnull):
            iraf.phot(f"{fits_path}[{extension}]",
                      coords=daofind_file.name,
                      output=phot_file.name,
                      ParList=phot_params_file.name)

        # Filter undefined magnitudes and export a well-formatted CSV file
        iraf.txselect(phot_file.name, filtered_phot_file.name, "MAG!=INDEF && MERR < 2.")
        iraf.txdump(filtered_phot_file.name,
                    fields="XC,YC,MAG,MERR", expr="yes", header="no", parameters="yes",
                    Stdout=csv_path)
        with open(csv_path, "r+") as output_file:
            rows = output_file.read().replace("  ", ",")
            output_file.seek(0)
            output_file.truncate(0)
            output_file.write("x,y,magnitude,magnitude_error\n")
            output_file.write(rows)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_fits_path", help="Path to the FITS image to be processed with iraf")
    parser.add_argument("output_csv_path", help="Path to the output CSV file to be generated")
    options = parser.parse_args()
    print(f"Processing {options.input_fits_path} into {options.output_csv_path}...")
    fits_to_csv(options.input_fits_path, options.output_csv_path)
    print(f"Done processing {options.input_fits_path} into {options.output_csv_path}!")



import enb.plugins


class JPEGPlugin(enb.plugins.PluginMake):
    name = "jpeg"
    label = "Reference JPEG and JPEG-LS implementation"
    tags = {"data compression", "image", "codec"}
    contrib_authors = ["Thomas Richter"]
    contrib_reference_urls = ["https://github.com/thorfdbg/libjpeg/"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/libjpeg-master-1.59.zip?raw=true",
         "libjpeg-master-1.59.zip")]
    tested_on = {"linux"}



import textwrap
import enb.plugins


class JPEGXLPlugin(enb.plugins.PluginMake):
    name = "jpegxl"
    label = "Wrapper for a reference JPEG-XL implementation"
    tags = {"data compression", "image", "codec"}
    contrib_authors = ["See https://gitlab.com/wg1/jpeg-xl/-/blob/main/CONTRIBUTORS"]
    contrib_reference_urls = ["https://gitlab.com/wg1/jpeg-xl.git"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/jpegxl_git_20241112.zip?raw=true",
         "jpegxl_git.zip")]
    extra_requirements_message = textwrap.dedent("""\
            The cmake tool is needed to build this plugin, if not present already. 
            You can install cmake as follows: 

            * Ubuntu: `sudo apt install cmake`
            * MacOS:
                1. Install homebrew from http://brew.sh
                2. `brew install gcc@9`
                
                
            Also, if execution fails due to the lack of libraries such as `libjxl_threads.so.0.8`, 
            you may need to manually compile the plugin and copy the resulting libraries, e.g., in `/usr/lib`.
            The generated dynamic libraries are copied into the installation directory. 
            """)
    tested_on = {"linux"}



import os
import shutil
import matplotlib
import enb
from enb.config import options


class MatplotlibrcTemplate(enb.plugins.Template):
    """Copy a full enb.ini configuration in the destination project folder.
    """
    name = "matplotlibrc"
    author = ["Miguel Hernández-Cabronero"]
    label = "Copy matplotlib's default rc file into the destination directory."
    tags = {"project"}
    tested_on = {"linux", "windows"}

    @classmethod
    def build(cls, installation_dir):
        output_path = os.path.join(installation_dir, "matplotlibrc")
        if os.path.exists(output_path) and not options.force:
            raise ValueError(f"Output file {output_path} exists. Refusing to overwrite.")
        shutil.copy(matplotlib.matplotlib_fname(), output_path)
        print("\nSee https://matplotlib.org/stable/tutorials/introductory/customizing.html#the-matplotlibrc-file \n"
              "for information on how to edit this file.\n")



import enb.plugins


class FapecPlugin(enb.plugins.Plugin):
    name = "fapec"
    label = "Wrappers for FAPEC"
    tags = {"data compression", "image", "codec", "privative"}
    contrib_authors = ["DAPCOM"]
    contrib_reference_urls = ["https://www.dapcom.es/fapec/"]
    tested_on = {"linux"}



import enb.plugins


class FPCPlugin(enb.plugins.PluginMake):
    name = "fpc"
    label = "FPC codec wrappers"
    tags = {"data compression", "codec"}
    contrib_authors = ["Martin Burtscher et al."]
    contrib_reference_urls = ["https://userweb.cs.txstate.edu/~burtscher/research/FPC/"]
    tested_on = {"linux"}



import enb.plugins


class LCNLPlugin(enb.plugins.Plugin):
    name = "lcnl"
    label = "Wrappers for LCNL CCSDS 123.0-B-2"
    tags = {"data compression", "image", "codec", "privative"}
    contrib_authors = ["Ian Blanes", "Miguel Hernández-Cabronero", "CNES, et al."]
    contrib_reference_urls = [
        "http://deic.uab.cat/~mhernandez/media/papers/Miguel_Hernandez-IEEE_GRSM_123x0b2_explained.pdf"]
    extra_requirements_message = "Low-complexity near-lossless image compression (LCNL), CCSDS 123.0-B2\n" \
                                 "The source code and the binaries are not redistributed with enb due to intellectual property issues.\n" \
                                 "The binaries can be obtained at https://www.connectbycnes.fr/en/ccsds-1230-b-2-ccsds-1210-b-3 provided by CNES."
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Data compression tools common to any compression modality.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/01"

import os
import re
import filecmp
import hashlib
import tempfile
import time
import collections
import shutil
import math
from typing import Optional

import numpy as np

from scipy import signal
from scipy.ndimage.filters import convolve

import enb
import enb.isets
import enb.tcall
from enb.config import options


class CompressionResults:
    """Base class that defines the minimal fields that are returned by a call
    to a coder's compress() method (or produced by the CompressionExperiment
    instance).
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, codec_name=None, codec_param_dict=None,
                 original_path=None,
                 compressed_path=None,
                 compression_time_seconds=None,
                 maximum_memory_kb=None):
        """
        :param codec_name: codec's reported_name
        :param codec_param_dict: dictionary of parameters to the codec
        :param original_path: path to the input original file
        :param compressed_path: path to the output compressed file
        :param compression_time_seconds: effective average compression time in
          seconds
        :param maximum_memory_kb: maximum resident memory in kilobytes
        """
        # pylint: disable=too-many-arguments
        self.codec_name = codec_name
        self.codec_param_dict = codec_param_dict
        self.original_path = original_path
        self.compressed_path = compressed_path
        self.compression_time_seconds = compression_time_seconds
        self.maximum_memory_kb = maximum_memory_kb


class DecompressionResults:
    """Base class that defines the minimal fields that are returned by a call
    to a coder's decompress() method (or produced by the
    CompressionExperiment instance).
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, codec_name=None, codec_param_dict=None,
                 compressed_path=None,
                 reconstructed_path=None,
                 decompression_time_seconds=None,
                 maximum_memory_kb=None):
        """
        :param codec_name: codec's reported_name
        :param codec_param_dict: dictionary of parameters to the codec
        :param compressed_path: path to the output compressed file
        :param reconstructed_path: path to the reconstructed file after
          decompression
        :param decompression_time_seconds: effective decompression time in
          seconds
        :param maximum_memory_kb: maximum resident memory in kilobytes
        """
        # pylint: disable=too-many-arguments
        self.codec_name = codec_name
        self.codec_param_dict = codec_param_dict
        self.compressed_path = compressed_path
        self.reconstructed_path = reconstructed_path
        self.decompression_time_seconds = decompression_time_seconds
        self.maximum_memory_kb = maximum_memory_kb


class CompressionException(Exception):
    """Base class for exceptions occurred during a compression instance
    """

    def __init__(self, original_path=None, compressed_path=None, file_info=None,
                 status=None, output=None):
        # pylint: disable=too-many-arguments
        super().__init__(f"status={status}, output={output}")
        self.original_path = original_path
        self.compressed_path = compressed_path
        self.file_info = file_info
        self.status = status
        self.output = output


class DecompressionException(Exception):
    """Base class for exceptions occurred during a decompression instance
    """

    def __init__(self, compressed_path=None, reconstructed_path=None,
                 file_info=None, status=None, output=None):
        # pylint: disable=too-many-arguments
        super().__init__(f"status={status}, output={output}")
        self.reconstructed_path = reconstructed_path
        self.compressed_path = compressed_path
        self.file_info = file_info
        self.status = status
        self.output = output


class CompressionExperiment(enb.experiment.Experiment):
    """This class allows seamless execution of compression experiments.

    In the functions decorated with @atable,column_function, the row argument
    contains two magic properties, compression_results and decompression_results.
    These give access to the :class:`CompressionResults` and :class:`DecompressionResults`
    instances resulting respectively from compressing and decompressing
    according to the row index parameters. The paths referenced in the compression
    and decompression results are valid while the row is being processed, and
    are disposed of afterwards.
    Also, the image_info_row attribute gives access to the image metainformation
    (e.g., geometry)
    """
    dataset_files_extension = "raw"
    default_file_properties_table_class = enb.isets.ImagePropertiesTable
    row_wrapper_column_name = "_codec_wrapper"

    class CompressionDecompressionWrapper:
        """This class is instantiated for each row of the table, and added to a temporary
        column row_wrapper_column_name. Column-setting methods can then access this wrapper,
        and in particular its `compression_results` and `decompression_results` properties,
        which will run compression and decompression at most once. This way, many columns can
        be defined independently without needing to compress and decompress for each one.
        """

        def __init__(self, file_path, codec, image_info_row,
                     reconstructed_copy_dir=None,
                     compressed_copy_dir=None):
            """
            :param file_path: path to the original image being compressed
            :param codec: AbstractCodec instance to be used for compression/decompression
            :param image_info_row: dict-like object with geometry and
              data type information about file_path
            :param reconstructed_copy_dir: if not None, a copy of the reconstructed images
              is stored, based on the class of codec.
            :param compressed_copy_dir: if not None, a copy of the compressed images
              is stored, based on the class of codec.
            """
            # pylint: disable=too-many-arguments
            self.file_path = file_path
            self.codec = codec
            self.image_info_row = image_info_row
            self._compression_results = None
            self._decompression_results = None
            self.compressed_copy_dir = compressed_copy_dir
            self.reconstructed_copy_dir = reconstructed_copy_dir

        @property
        def compression_results(self) -> CompressionResults:
            """Perform the actual compression experiment for the selected row.
            """
            if self._compression_results is None:
                os.makedirs(options.base_tmp_dir, exist_ok=True)
                # Assign a unique temporary path to the compressed file and avoid introducing spurious geometry tags
                tmp_compressed_path = None
                while ((tmp_compressed_path is None) or
                       (re.search(r"\d+x\d+x\d+.*\D\d+x\d+x\d+", os.path.basename(tmp_compressed_path)) is not None)):
                    fd, tmp_compressed_path = tempfile.mkstemp(
                        dir=options.base_tmp_dir,
                        prefix=f"compressed_{os.path.basename(self.file_path)}_")
                    os.close(fd)
                    os.remove(tmp_compressed_path)

                try:
                    measured_times = []
                    measured_memory = []

                    enb.logger.debug(
                        f"Executing compression {self.codec.name} on {self.file_path} "
                        f"[{options.repetitions} times]")
                    for repetition_index in range(options.repetitions):
                        enb.logger.debug(
                            f"Executing compression {self.codec.name} on {self.file_path} "
                            f"[rep{repetition_index + 1}/{options.repetitions}]")
                        time_before_ns = time.time_ns()
                        self._compression_results = self.codec.compress(
                            original_path=self.file_path,
                            compressed_path=tmp_compressed_path,
                            original_file_info=self.image_info_row)

                        if not os.path.isfile(tmp_compressed_path) \
                                or os.path.getsize(tmp_compressed_path) == 0:
                            raise CompressionException(
                                original_path=self.file_path,
                                compressed_path=tmp_compressed_path,
                                file_info=self.image_info_row,
                                output=f"Compression of {self.file_path} "
                                       f"didn't produce a file (or it was empty)")

                        wall_compression_time = \
                            (time.time_ns() - time_before_ns) / 1e9
                        if self._compression_results is None:
                            enb.logger.debug(
                                f"[W]arning: codec {self.codec.name} "
                                f"did not report execution times. "
                                f"Using wall clock instead (might be inaccurate)")
                            self._compression_results = \
                                self.codec.compression_results_from_paths(
                                    original_path=self.file_path,
                                    compressed_path=tmp_compressed_path)
                            self._compression_results.compression_time_seconds = \
                                wall_compression_time

                        measured_times.append(
                            self._compression_results.compression_time_seconds)
                        measured_memory.append(
                            self._compression_results.maximum_memory_kb)

                        if self.compressed_copy_dir and repetition_index == 0:
                            output_path = os.path.join(
                                self.compressed_copy_dir,
                                self.codec.name,
                                f"{os.path.basename(self.file_path)}.compressed")
                            os.makedirs(os.path.dirname(output_path),
                                        exist_ok=True)
                            enb.logger.debug(
                                f"Storing compressed bitstream for "
                                f"{self.file_path} and {self.codec} "
                                f"at {repr(output_path)}")
                            shutil.copyfile(tmp_compressed_path, output_path)

                        if repetition_index < options.repetitions - 1:
                            os.remove(tmp_compressed_path)

                    # The minimum time is kept, all other values are
                    # considered to have noise added by the OS
                    self._compression_results.compression_time_seconds = min(
                        measured_times)
                    # The maximum resident memory in kb is kept
                    self._compression_results.maximum_memory_kb \
                        = max(kb if kb is not None else -1
                              for kb in measured_memory)
                except Exception as ex:
                    if os.path.exists(tmp_compressed_path):
                        os.remove(tmp_compressed_path)
                    raise ex

            return self._compression_results

        @property
        def decompression_results(self) -> DecompressionResults:
            """Perform the actual decompression experiment for the selected row.
            """
            if self._decompression_results is None:
                # Assign a unique temporary path to the reconstructed file and avoid introducing spurious geometry tags
                tmp_reconstructed_path = None
                while ((tmp_reconstructed_path is None) or
                       (re.search(r"\d+x\d+x\d+.*\D\d+x\d+x\d+",
                                  os.path.basename(tmp_reconstructed_path)) is not None)):
                    fd, tmp_reconstructed_path = tempfile.mkstemp(
                        dir=options.base_tmp_dir,
                        prefix=f"compressed_{os.path.basename(self.file_path)}_")
                    os.close(fd)
                    os.remove(tmp_reconstructed_path)
                    
                try:
                    measured_times = []
                    measured_memory = []
                    with enb.logger.debug_context(
                            f"Executing decompression {self.codec.name} "
                            f"on {self.file_path} [{options.repetitions} times]"
                            + ("\n" if options.repetitions > 1 else "")):
                        for repetition_index in range(options.repetitions):
                            enb.logger.debug(
                                f"Executing decompression {self.codec.name} "
                                f"on {self.file_path} "
                                f"[rep{repetition_index + 1}/{options.repetitions}]")

                            time_before = time.time_ns()
                            self._decompression_results = self.codec.decompress(
                                compressed_path=self.compression_results.compressed_path,
                                reconstructed_path=tmp_reconstructed_path,
                                original_file_info=self.image_info_row)

                            wall_decompression_time = \
                                (time.time_ns() - time_before) / 1e9
                            if self._decompression_results is None:
                                enb.logger.debug(
                                    f"Codec {self.codec.name} did not report "
                                    f"execution times. Using wall clock "
                                    f"instead (might be inaccurate with "
                                    f"intensive I/O)")
                                self._decompression_results = \
                                    self.codec.decompression_results_from_paths(
                                        compressed_path=self.compression_results.compressed_path,
                                        reconstructed_path=tmp_reconstructed_path)
                                self._decompression_results.decompression_time_seconds = \
                                    wall_decompression_time

                            if not os.path.isfile(
                                    tmp_reconstructed_path) or os.path.getsize(
                                self._decompression_results.reconstructed_path) == 0:
                                raise CompressionException(
                                    original_path=self.compression_results.original_path,
                                    compressed_path=self.compression_results.compressed_path,
                                    file_info=self.image_info_row,
                                    output=f"Decompression didn't produce a file (or it was empty)"
                                           f" {self.compression_results.original_path}")

                            measured_times.append(
                                self._decompression_results.decompression_time_seconds)
                            measured_memory.append(
                                self._decompression_results.maximum_memory_kb)

                            if self.reconstructed_copy_dir and repetition_index == 0:
                                output_path = os.path.join(
                                    self.reconstructed_copy_dir,
                                    self.codec.name,
                                    os.path.basename(self.file_path))
                                os.makedirs(os.path.dirname(output_path),
                                            exist_ok=True)
                                enb.logger.debug(
                                    f"Storing reconstructed copy of "
                                    f"{self.file_path} with {self.codec} "
                                    f"at {repr(output_path)}")
                                shutil.copyfile(tmp_reconstructed_path,
                                                output_path)

                            if repetition_index < options.repetitions - 1:
                                os.remove(tmp_reconstructed_path)
                    # The minimum time is kept, the remaining values are
                    # assumed to contain noised added by the OS
                    self._decompression_results.decompression_time_seconds = \
                        min(measured_times)
                    self._decompression_results.maximum_memory_kb \
                        = max(kb if kb is not None else -1
                              for kb in measured_memory)
                except Exception as ex:
                    os.remove(tmp_reconstructed_path)
                    raise ex

            return self._decompression_results

        @property
        def numpy_dtype(self):
            """Get the numpy dtype corresponding to the original image's data
            format
            """

            return enb.isets.iproperties_row_to_numpy_dtype(
                self.image_info_row)

        def __del__(self):
            if self._compression_results is not None:
                try:
                    os.remove(self._compression_results.compressed_path)
                except OSError:
                    pass
            if self._decompression_results is not None:
                try:
                    os.remove(self._decompression_results.reconstructed_path)
                except OSError:
                    pass

    def __init__(self, codecs,
                 dataset_paths=None,
                 csv_experiment_path=None,
                 csv_dataset_path=None,
                 dataset_info_table=None,
                 overwrite_file_properties=False,
                 reconstructed_dir_path=None,
                 compressed_copy_dir_path=None,
                 task_families=None):
        """
        :param codecs: list of :py:class:`AbstractCodec` instances. Note that
          codecs are compatible with the interface of :py:class:`ExperimentTask`.
        :param dataset_paths: list of paths to the files to be used as input
          for compression. If it is None, this list is obtained automatically
          from the configured base dataset dir.
        :param csv_experiment_path: if not None, path to the CSV file giving
          persistence support to this experiment. If None, it is automatically
          determined within options.persistence_dir.
        :param csv_dataset_path: if not None, path to the CSV file given
          persistence support to the dataset file properties. If None,
          it is automatically determined within options.persistence_dir.
        :param dataset_info_table: if not None, it must be a
          ImagePropertiesTable instance or subclass instance that can be used
          to obtain dataset file metainformation, and/or gather it from
          csv_dataset_path. If None, a new ImagePropertiesTable instance is
          created and used for this purpose.
        :param overwrite_file_properties: if True, file properties are
          recomputed before starting the experiment. Useful for temporary
          and/or random datasets. Note that overwrite control for the
          experiment results themselves is controlled in the call to get_df
        :param reconstructed_dir_path: if not None, a directory where
          reconstructed images are to be stored.
        :param compressed_copy_dir_path: if not None, it gives the directory
          where a copy of the compressed images. is to be stored. If may not be
          generated for images for which all columns are known
        :param task_families: if not None, it must be a list of TaskFamily
          instances. It is used to set the "family_label" column for each row.
          If the codec is not found within the families, a default label is set
          indicating so.
        """
        # pylint: disable=too-many-arguments
        table_class = type(dataset_info_table) if dataset_info_table is not None \
            else self.default_file_properties_table_class
        csv_dataset_path = csv_dataset_path if csv_dataset_path is not None \
            else os.path.join(options.persistence_dir,
                              f"persistence_{table_class.__name__}.csv")
        imageinfo_table = dataset_info_table if dataset_info_table is not None \
            else table_class(csv_support_path=csv_dataset_path)

        csv_dataset_path = csv_dataset_path if csv_dataset_path is not None \
            else f"persistence_{dataset_info_table.__class__.__name__}.csv"

        if not codecs:
            raise ValueError(
                f"{self}: no codecs were selected for this experiment. At least one is needed.")
        non_subclass_codecs = [c for c in codecs if not isinstance(c, enb.compression.codec.AbstractCodec)]
        if non_subclass_codecs:
            enb.logger.warn(
                f"Compression experiment {self.__class__.__name__} "
                f"received parameter codecs "
                f"with {len(non_subclass_codecs)} objects not inheriting from "
                f"enb.icompression.AbstractCodec: "
                f"{', '.join(repr(c) for c in non_subclass_codecs)}.\n"
                f"You can remove this warning by explicitly inheriting "
                f"from that class for "
                f"the aforementioned instances.")

        super().__init__(tasks=codecs,
                         dataset_paths=dataset_paths,
                         csv_experiment_path=csv_experiment_path,
                         csv_dataset_path=csv_dataset_path,
                         dataset_info_table=imageinfo_table,
                         overwrite_file_properties=overwrite_file_properties,
                         task_families=task_families)
        self.reconstructed_dir_path = reconstructed_dir_path
        self.compressed_copy_dir_path = compressed_copy_dir_path
        # This attribute is automatically set before running the defined
        # column-setting functions, then set back to None after that. It
        # enables lazy and at-most-once compression/decompression.
        self.codec_results: Optional[CompressionExperiment.CompressionDecompressionWrapper] = None

    @property
    def codecs(self):
        """:return: an iterable of defined codecs
        """
        return self.tasks_by_name.values()

    @codecs.setter
    def codecs(self, new_codecs):
        self.tasks_by_name = collections.OrderedDict({
            codec.name: codec for codec in new_codecs})

    @property
    def codecs_by_name(self):
        """Alias for :py:attr:`tasks_by_name`
        """
        return self.tasks_by_name

    @property
    def compression_results(self) -> CompressionResults:
        """Get the current compression results from self.codec_results.
        This property is intended to be read from functions that set columns of a row.
        It triggers the compression of that row's sample with that row's codec if it hasn't been compressed yet.
        Otherwise, None is returned.
        """
        return self.codec_results.compression_results if self.codec_results else None

    @property
    def decompression_results(self) -> DecompressionResults:
        """Get the current decompression results from self.codec_results.
        This property is intended to be read from functions that set columns of a row.
        It triggers the compression and decompression of that row's sample with that row's codec if
        they have not been compressed yet.
        Otherwise, None is returned.
        """
        return self.codec_results.decompression_results if self.codec_results else None

    def compute_one_row(self, filtered_df, index, loc, column_fun_tuples, overwrite):
        # pylint: disable=too-many-arguments
        # Prepare a new column with a self.CodecRowWrapper instance that
        # allows automatic, lazy computation of compression and decompression
        # results.
        file_path, codec_name = index
        codec = self.codecs_by_name[codec_name]
        image_info_row = self.dataset_table_df.loc[
            enb.atable.indices_to_internal_loc(file_path)]

        # A temporary attribute is created with a
        # self.CompressionDecompressionWrapper instance, which allows lazy,
        # at-most-one execution of the compression/decompression process.
        # Column-setting methods can access the wrapper with self.
        try:
            assert self.codec_results is None
        except AttributeError:
            pass
        try:
            self.codec_results = self.CompressionDecompressionWrapper(
                file_path=file_path, codec=codec,
                image_info_row=image_info_row,
                compressed_copy_dir=self.compressed_copy_dir_path or enb.config.options.compressed_copy_dir,
                reconstructed_copy_dir=self.reconstructed_dir_path or enb.config.options.reconstructed_dir)
            assert self.codec_results is not None

            processed_row = super().compute_one_row(
                filtered_df=filtered_df, index=index, loc=loc,
                column_fun_tuples=column_fun_tuples,
                overwrite=overwrite)

            if isinstance(processed_row, Exception):
                # Should not do anything beyond here if errors occurred
                return processed_row
        finally:
            del self.codec_results
            self.codec_results = None

            try:
                del self.codec_results
            except AttributeError:
                pass

        return processed_row

    @enb.atable.column_function("compressed_size_bytes",
                                label="Compressed data size (Bytes)",
                                plot_min=0)
    def set_compressed_data_size(self, index, row):
        row[_column_name] = os.path.getsize(
            self.codec_results.compression_results.compressed_path)

    @enb.atable.column_function([
        enb.atable.ColumnProperties(name="compression_ratio",
                                    label="Compression ratio", plot_min=0),
        enb.atable.ColumnProperties(name="lossless_reconstruction",
                                    label="Lossless?"),
        enb.atable.ColumnProperties(name="compression_time_seconds",
                                    label="Compression time (s)", plot_min=0),
        enb.atable.ColumnProperties(name="decompression_time_seconds",
                                    label="Decompression time (s)", plot_min=0),
        enb.atable.ColumnProperties(name="repetitions",
                                    label="Number of compression/decompression "
                                          "repetitions",
                                    plot_min=0),
        enb.atable.ColumnProperties(name="compressed_file_sha256",
                                    label="Compressed file's SHA256"),
        enb.atable.ColumnProperties(name="compression_memory_kb",
                                    label="Compression memory usage (KB)",
                                    plot_min=0),
        enb.atable.ColumnProperties(name="decompression_memory_kb",
                                    label="Decompression memory usage (KB)",
                                    plot_min=0),
    ])
    def set_comparison_results(self, index, row):
        """Perform a compression-decompression cycle and store the comparison
        results
        """
        file_path, codec_name = self.index_to_path_task(index)
        row.image_info_row = self.dataset_table_df.loc[
            enb.atable.indices_to_internal_loc(file_path)]
        assert self.codec_results.compression_results.compressed_path \
               == self.codec_results.decompression_results.compressed_path
        try:
            assert row.image_info_row["bytes_per_sample"] * row.image_info_row["samples"] \
                   == os.path.getsize(self.codec_results.compression_results.original_path)
        except (KeyError, AssertionError) as ex:
            enb.logger.debug(f"Could not verify valid size. {repr(ex)}")
        hasher = hashlib.sha256()
        with open(self.codec_results.compression_results.compressed_path,
                  "rb") as compressed_file:
            hasher.update(compressed_file.read())
        compressed_file_sha256 = hasher.hexdigest()

        row["lossless_reconstruction"] = filecmp.cmp(
            self.codec_results.compression_results.original_path,
            self.codec_results.decompression_results.reconstructed_path)
        assert self.codec_results.compression_results.compression_time_seconds \
               is not None
        row["compression_time_seconds"] = \
            self.codec_results.compression_results.compression_time_seconds
        assert self.codec_results.decompression_results.decompression_time_seconds \
               is not None
        row["decompression_time_seconds"] = \
            self.codec_results.decompression_results.decompression_time_seconds
        row["repetitions"] = options.repetitions
        row["compression_ratio"] = os.path.getsize(
            self.codec_results.compression_results.original_path) \
                                   / row["compressed_size_bytes"]
        row["compressed_file_sha256"] = compressed_file_sha256
        row["compression_memory_kb"] = \
            self.codec_results.compression_results.maximum_memory_kb
        row["decompression_memory_kb"] = \
            self.codec_results.decompression_results.maximum_memory_kb

    @enb.atable.column_function("bpppc", label="Compressed data rate (bpppc)",
                                plot_min=0)
    def set_bpppc(self, index, row):
        file_path, codec_name = self.index_to_path_task(index)
        row.image_info_row = self.dataset_table_df.loc[
            enb.atable.indices_to_internal_loc(file_path)]
        try:
            row[_column_name] = 8 * row["compressed_size_bytes"] / \
                                row.image_info_row["samples"]
        except KeyError as ex:
            enb.logger.debug(f"Cannot determine bpppc: {repr(ex)}")
            assert "compressed_size_bytes" in row

    @enb.atable.column_function("compression_ratio_dr",
                                label="Compression ratio",
                                plot_min=0)
    def set_compression_ratio_dr(self, index, row):
        """Set the compression ratio calculated based on the dynamic range of
        the input samples, as opposed to 8*bytes_per_sample.
        """
        file_path, codec_name = self.index_to_path_task(index)
        row.image_info_row = self.dataset_table_df.loc[
            enb.atable.indices_to_internal_loc(file_path)]
        row[_column_name] = (row.image_info_row["dynamic_range_bits"] *
                             row.image_info_row["samples"]) \
                            / (8 * row["compressed_size_bytes"])

    @enb.atable.column_function([
        enb.atable.ColumnProperties(
            name=f"compression_efficiency_{B}byte_entropy",
            labytesel=f"Compression efficiency ({B}bytes entropy)",
            plot_min=0, plot_max=B) for B in (1, 2, 4)])
    def set_efficiency(self, index, row):
        file_path, codec_name = self.index_to_path_task(index)
        row.image_info_row = self.dataset_table_df.loc[
            enb.atable.indices_to_internal_loc(file_path)]
        for B in (1, 2, 4):
            column_name = f"compression_efficiency_{B}byte_entropy"
            try:
                row[column_name] = \
                    (row.image_info_row[f"entropy_{B}B_bps"]
                     * (row.image_info_row["size_bytes"] / B)
                     / (row["compressed_size_bytes"] * 8))
                if row[column_name] < 0:
                    # The entropy is not known (e.g., size not multiple of B bytes)
                    row[column_name] = -1
            except KeyError as ex:
                enb.logger.warn(
                    f"Could not find a column required to compute "
                    f"{column_name}: {repr(ex)}. "
                    f"Setting to -1. This is likely due to persistence "
                    f"data produced with "
                    f"version v0.4.2 or older of enb. It is recommended "
                    f"to delete persistence "
                    f"data files and re-run the experiment.")
                row[column_name] = -1


class LosslessCompressionExperiment(CompressionExperiment):
    """Lossless data compression experiment. It fails if lossless reconstruction is not achieved.
    """

    @enb.atable.redefines_column
    def set_comparison_results(self, index, row):
        path, task = self.index_to_path_task(index)
        super().set_comparison_results(index=index, row=row)
        if not row["lossless_reconstruction"]:
            raise CompressionException(
                original_path=index[0], file_info=row,
                output="Failed to produce lossless compression for "
                       f"{path} and {task}")


class GenericFilePropertiesTable(enb.isets.ImagePropertiesTable):
    """File properties table that considers the input path as a 1D,
    u8be array.
    """
    verify_file_size = False

    @enb.atable.column_function([
        enb.atable.ColumnProperties(name="unique_sample_count",
                                    label="Number of different sample values"),
        enb.atable.ColumnProperties(name="sample_min",
                                    label="Min sample value (byte samples)"),
        enb.atable.ColumnProperties(name="sample_max",
                                    label="Max sample value (byte samples)")])
    def set_sample_stats(self, file_path, row):
        """Set basic file statistics (unique count, min, max)
        """
        with open(file_path, "rb") as input_file:
            contents = set(input_file.read())
            row["unique_sample_count"] = len(contents)
            row["sample_min"] = min(contents)
            row["sample_max"] = max(contents)

    @enb.atable.column_function("bytes_per_sample",
                                label="Bytes per sample",
                                plot_min=0)
    def set_bytes_per_sample(self, file_path, row):
        row[_column_name] = 1

    @enb.atable.column_function([
        enb.atable.ColumnProperties(name="width", label="Width",
                                    plot_min=1),
        enb.atable.ColumnProperties(name="height", label="Height",
                                    plot_min=1),
        enb.atable.ColumnProperties(name="component_count",
                                    label="Components",
                                    plot_min=1),
        enb.atable.ColumnProperties(name="big_endian"),
        enb.atable.ColumnProperties(name="float"),
    ])
    def set_image_geometry(self, file_path, row):
        """Obtain the image's geometry (width, height and number of
        components) based on the filename tags (and possibly its size)
        """
        row["height"] = 1
        row["width"] = os.path.getsize(file_path)
        row["component_count"] = 1
        row["big_endian"] = True
        row["float"] = False


class GeneralLosslessExperiment(LosslessCompressionExperiment):
    """Lossless compression experiment for general data contents.
    """
    default_file_properties_table_class = GenericFilePropertiesTable
    dataset_files_extension = ""



#!/usr/bin/env python3
"""Template class for transformation using mhdctransform
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/05/25"

import os
import tempfile
import filecmp
import shutil

import enb.atable
from enb.config import get_options

options = get_options(from_main=False)

import enb
import enb.sets
import enb.isets
import enb.parallel_ray
import enb.tcall
import enb.atable as atable

default_mhdc_binary_path = os.path.join(os.path.dirname(__file__), "mhdctransform")


def output_path_to_si_path(output_path, transform_number):
    return f"{output_path}_SI.mhdc_t{transform_number}"


class MHDCGeometryTable(enb.isets.ImageGeometryTable):
    @enb.atable.redefines_column
    def set_signed(self, file_path, row):
        """"""
        row[_column_name] = True


class MHDCPropertiesTable(MHDCGeometryTable, enb.isets.ImagePropertiesTable):
    """Properties table for forward MHDC-transformed images"""
    pass


class MHDCTransformTable(enb.sets.FileVersionTable, enb.isets.ImagePropertiesTable):
    mhdc_transform_path = default_mhdc_binary_path

    invocation_output_folder = None

    @property
    def transform_number(self):
        raise NotImplementedError("Subclasses must return the appropriate transform number")

    @property
    def compacted_bytes_per_sample(self):
        """ Number of bytes per sample used in the final transformed files.
        If None is returned, the nominal bytes per sample of the original data
        are maintained.
        """
        return None

    def get_specific_params(self, input_path, output_path):
        """Return a string of parameters specific to this transform,
        or an empty string if none are needed.
        """
        return ""

    def compact_samples(self, transformed_path, compacted_path, row):
        """The mhdctransform binary always outputs in 4-bytes-per-sample, signed format.
        This method compacts this information into s16b.

        :param transformed_path: path to the data in s32b
        :param compacted_path: path to the data in s16b (can be the same as trasformed_path,
          in that case it is overwriten)
        """
        input_row = row.copy()
        input_row["bytes_per_sample"] = 4
        input_row["signed"] = True
        s32b_array = enb.isets.load_array_bsq(file_or_path=transformed_path, image_properties_row=input_row)

        output_row = row.copy()
        output_row["signed"] = True
        if self.compacted_bytes_per_sample is not None:
            output_row["bytes_per_sample"] = self.compacted_bytes_per_sample
        output_dtype = enb.isets.iproperties_row_to_numpy_dtype(output_row)
        enb.isets.dump_array_bsq(array=s32b_array, file_or_path=compacted_path, dtype=output_dtype)

    def version(self, input_path, output_path, row):
        assert os.path.exists(self.mhdc_transform_path), \
            f"mhdc transform binary not found at '{self.mhdc_transform_path}' - please reset and retry"
        invocation = f"/usr/bin/time -v {self.mhdc_transform_path} " \
                     f"--Size {row['component_count']} {row['height']} {row['width']} " \
                     f"--Bitdepth {8 * row['bytes_per_sample']} " \
                     f"--SampleType {1 if row['signed'] else 0} " \
                     f"--SpectralTransform {self.transform_number} --Direction 0 " \
                     f"{self.get_specific_params(input_path=input_path, output_path=output_path)} " \
                     f"{input_path} " \
                     f"{output_path_to_si_path(output_path=output_path, transform_number=self.transform_number)} " \
                     f"{output_path}"
        status, output, reported_time = enb.tcall.get_status_output_time(invocation)

        # s32b is hardcoded in mhdctransform - revert to original bytes per sample
        self.compact_samples(transformed_path=output_path, compacted_path=output_path, row=row)

        if self.invocation_output_folder is not None:
            os.makedirs(self.invocation_output_folder, exist_ok=True)
            invocation_output_path = os.path.join(
                self.invocation_output_folder,
                self.__class__.__name__ + "_" + os.path.abspath(os.path.realpath(input_path)).replace(os.sep, "_"))
            with open(invocation_output_path, "w") as invocation_output_file:
                invocation_output_file.write(f"Input_path: {input_path}\n "
                                             f"Invocation: {invocation}\n "
                                             f"Status: {status}\n "
                                             f"Output: {output}")

        return reported_time

    @atable.redefines_column
    def set_signed(self, file_path, row):
        """MHDC-transformed images are always signed
        """
        row[_column_name] = True


class InverseMHDCTransformTable(enb.sets.FileVersionTable, enb.isets.ImageGeometryTable):
    """VersionTable that applies an inverse MHDC transform and verifies that the result
    is lossless.
    """
    mhdc_transform_path = default_mhdc_binary_path

    @property
    def transform_number(self):
        raise NotImplementedError("Subclasses must return the appropriate transform number "
                                  "(should be the same as for the Direct version")

    @property
    def compacted_bytes_per_sample(self):
        """ Number of bytes per sample used in the final transformed files.
        If None is returned, the nominal bytes per sample of the original data
        are maintained.
        """
        return None

    def expand_sample(self, compacted_path, expanded_path, original_row):
        """Apply the inverse of MHDCTransformTable.compact_samples, i.e.,
        transform data into s32b format.
        """
        input_row = original_row.copy()
        input_row["signed"] = True
        array = enb.isets.load_array_bsq(file_or_path=compacted_path, image_properties_row=input_row)
        output_row = original_row.copy()
        output_row["bytes_per_sample"] = 4
        output_row["signed"] = True
        hardcoded_type = enb.isets.iproperties_row_to_numpy_dtype(image_properties_row=output_row)
        enb.isets.dump_array_bsq(array=array, file_or_path=expanded_path, dtype=hardcoded_type)

    def version(self, input_path, output_path, row):
        si_path = output_path_to_si_path(output_path=input_path, transform_number=self.transform_number)
        assert os.path.exists(si_path), f"Side information {si_path} does not exist but is necessary."

        with tempfile.NamedTemporaryFile(dir=options.base_tmp_dir) as tmp_input:
            original_row = row.copy()

            if self.compacted_bytes_per_sample is None:
                enb.isets.ImageGeometryTable.set_bytes_per_sample.__globals__.update(_column_name="bytes_per_sample")
                enb.isets.ImageGeometryTable.set_bytes_per_sample(
                    self.original_properties_table, file_path=input_path, row=original_row)
                del enb.isets.ImageGeometryTable.set_bytes_per_sample.__globals__["_column_name"]
            else:
                original_row["bytes_per_sample"] = self.compacted_bytes_per_sample
            enb.isets.ImageGeometryTable.set_signed.__globals__.update(_column_name="signed")
            enb.isets.ImageGeometryTable.set_signed(
                self.original_properties_table, file_path=input_path, row=original_row)
            del enb.isets.ImageGeometryTable.set_signed.__globals__["_column_name"]

            # Transform back to hardcoded s32b
            self.expand_sample(compacted_path=input_path, expanded_path=tmp_input.name,
                               original_row=original_row)

            invocation = f"{self.mhdc_transform_path} " \
                         f"--Size {original_row['component_count']} {original_row['height']} {original_row['width']} " \
                         f"--Bitdepth {8 * original_row['bytes_per_sample']} " \
                         f"--SampleType {1 if original_row['signed'] else 0} " \
                         f"--SpectralTransform {self.transform_number} --Direction 1 " \
                         f"{tmp_input.name} " \
                         f"{output_path_to_si_path(output_path=input_path, transform_number=self.transform_number)} " \
                         f"{output_path}"
            status, output, invocation_time = enb.tcall.get_status_output_time(invocation)
            return invocation_time


class MDHCLosslessCompressionExperiment(enb.icompression.LosslessCompressionExperiment):
    def __init__(self, codecs, dataset_paths=None, csv_experiment_path=None, csv_dataset_path=None,
                 overwrite_file_properties=False, reconstructed_dir_path=None, *args, **kwargs):
        dataset_info_table = MHDCPropertiesTable(
            csv_support_path=csv_dataset_path)

        super().__init__(codecs=codecs, dataset_paths=dataset_paths, csv_experiment_path=csv_experiment_path,
                         csv_dataset_path=csv_dataset_path, dataset_info_table=dataset_info_table,
                         overwrite_file_properties=overwrite_file_properties,
                         reconstructed_dir_path=reconstructed_dir_path,
                         *args, **kwargs)

    @property
    def transform_number(self):
        raise NotImplementedError()

    @enb.atable.redefines_column
    def set_compressed_data_size(self, index, row):
        si_path = output_path_to_si_path(
            output_path=row[enb.sets.FilePropertiesTable.index_name],
            transform_number=self.transform_number)
        assert os.path.exists(si_path), f"Side information {si_path} not found. Is it really a RWA-transformed set?"
        row[_column_name] = os.path.getsize(self.codec_results.compression_results.compressed_path) \
                            + os.path.getsize(si_path)


def apply_transform(
        input_dir, output_dir, forward_class, inverse_class,
        forward_properties_csv=None, inverse_properties_csv=None,
        repetitions=10):
    """Apply an MHDC transform to input_dir and save the results to output_dir.

    :param input_dir: input directory to be transformed (recursively)
    :param output_dir: path to the output transformed dir
    :param forward_properties_csv: if not None, properties of the transformed images
      are stored here (including fwd transform time)
    :param inverse_properties_csv: if not None, properties of the reconstructed images
      are stored here (including inverse transform time)
    :param repetitions: number of repetitions used to calculate execution time
      (for both forward and inverse transform)

    :raises AssertionError: if transformation/reconstruction is not lossless

    :return: fwd_df, inv_df, i.e., the dataframes obtained with get_df
      for forward_class and inverse_class, respectively.
    """
    assert repetitions >= 1

    with tempfile.TemporaryDirectory(dir=options.base_tmp_dir) as tmp_dir:
        assert os.path.isdir(input_dir)
        os.makedirs(output_dir, exist_ok=True)

        options.persistence_dir = tmp_dir
        options.base_dataset_dir = input_dir
        options.base_version_dataset_dir = output_dir
        options.repetitions = repetitions
        reconstructed_dir_path = os.path.join(tmp_dir, "reconstructed_dir")
        os.makedirs(reconstructed_dir_path, exist_ok=True)

        enb.parallel_ray.init_ray()

        original_target_files = enb.atable.get_all_input_files()

        if options.verbose:
            print(f"[F]orward MHDC<{forward_class.__name__}> to {len(original_target_files)} images")
        original_geometry_table = enb.isets.ImageGeometryTable(
            csv_support_path=os.path.join(options.persistence_dir, "original_properties.csv"))
        forward_table = forward_class(
            original_base_dir=options.base_dataset_dir, version_base_dir=options.base_version_dataset_dir,
            original_properties_table=original_geometry_table, version_name=forward_class.__name__,
            csv_support_path=os.path.join(options.persistence_dir, "versioned_properties.csv"))
        forward_df = forward_table.get_df(target_indices=original_target_files, overwrite=options.force)

        if options.verbose:
            print(f"[I]nverse MHDC<{inverse_class.__name__}> to {len(original_target_files)} images")
        transformed_target_files = enb.atable.get_all_input_files(base_dataset_dir=options.base_version_dataset_dir)
        with tempfile.NamedTemporaryFile(dir=tmp_dir) as tmp_csv_support:
            shutil.copy(forward_table.csv_support_path, tmp_csv_support.name)
            fwd_ig_table = enb.isets.ImageGeometryTable(csv_support_path=tmp_csv_support.name)
            inverse_table = inverse_class(
                original_base_dir=options.base_version_dataset_dir,
                version_base_dir=reconstructed_dir_path,
                original_properties_table=fwd_ig_table,
                version_name=inverse_class.__name__,
                csv_support_path=os.path.join(
                    options.persistence_dir, "inverse_versioned_properties.csv"))
            inverse_df = inverse_table.get_df(target_indices=transformed_target_files)

        if options.verbose:
            print(f"[C]hecking lossless to {len(original_target_files)}  images")
        checked_count = 0
        for original_path in original_target_files:
            reconstructed_path = os.path.abspath(os.path.realpath(original_path)).replace(
                options.base_dataset_dir, reconstructed_dir_path)
            assert filecmp.cmp(original_path, reconstructed_path), \
                f"Not lossless transform {original_path}<->{reconstructed_path}"
            checked_count += 1
        assert checked_count == len(original_target_files)

        if forward_properties_csv is not None:
            os.makedirs(os.path.dirname(os.path.abspath(forward_properties_csv)), exist_ok=True)
            shutil.copyfile(forward_table.csv_support_path, forward_properties_csv)
        if inverse_properties_csv is not None:
            os.makedirs(os.path.dirname(os.path.abspath(inverse_properties_csv)), exist_ok=True)
            shutil.copyfile(inverse_table.csv_support_path, inverse_properties_csv)

        return forward_df, inverse_df



"""Implementation of a codec that applies a bitshuffle transform before compressing with a codec instance
provided to the constructor.
"""
import os
import tempfile
import time

import bitshuffle
import numpy as np

import enb


class BitshuffleWrapper(enb.icompression.AbstractCodec):
    """Codec that applies bitshuffle to the data before compressing with a codec instance
    provided to the constructor.
    """

    def __init__(self, codec: enb.icompression.AbstractCodec):
        """
        :param codec: The codec instance used to compress and decompress the transformed data.
        """
        super().__init__(param_dict=dict(codec_name=codec.name))
        self.codec = codec

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        """Apply bitshuffle and then compress with `self.codec`.
        """
        with tempfile.TemporaryDirectory(dir=enb.config.options.base_tmp_dir) as tmp_dir:
            # Save the transformed data to a temporary file (likely memory)
            shuffled_path = os.path.join(tmp_dir, os.path.basename(original_path))
            original_data = enb.isets.load_array(
                file_or_path=original_path,
                image_properties_row=original_file_info)
            time_before = time.process_time() if not enb.config.options.report_wall_time else time.time()
            shuffled_array = bitshuffle.bitshuffle(np.ascontiguousarray(original_data))
            time_after = time.process_time() if not enb.config.options.report_wall_time else time.time()
            shuffle_time = max(0.0, time_after - time_before)
            enb.isets.dump_array(array=shuffled_array, file_or_path=shuffled_path)

            # Compress the transformed data normally
            time_before = time.process_time() if not enb.config.options.report_wall_time else time.time()
            compression_results = self.codec.compress(
                original_path=shuffled_path,
                compressed_path=compressed_path,
                original_file_info=original_file_info
            )
            time_after = time.process_time() if not enb.config.options.report_wall_time else time.time()
            compression_time = max(0.0, time_after - time_before)
            if compression_results is None:
                compression_results = self.compression_results_from_paths(
                    original_path=shuffled_path,
                    compressed_path=compressed_path)
            os.remove(shuffled_path)

            compression_results.original_path = original_path
            if compression_results.compression_time_seconds is None:
                compression_results.compression_time_seconds = compression_time
            compression_results.compression_time_seconds += shuffle_time

            return compression_results

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        """Decompress with `self.codec` and then revert the bitshuffle transform.
        """
        with tempfile.TemporaryDirectory(dir=enb.config.options.base_tmp_dir) as tmp_dir:
            shuffled_path = os.path.join(tmp_dir, f"{reconstructed_path}_shuffled")

            # Decompress the shuffled data
            time_before = time.process_time() if not enb.config.options.report_wall_time else time.time()
            decompression_results = self.codec.decompress(
                compressed_path=compressed_path,
                reconstructed_path=shuffled_path)
            time_after = time.process_time() if not enb.config.options.report_wall_time else time.time()
            decompression_time = max(0.0, time_after - time_before)
            if decompression_results is None:
                decompression_results = self.decompression_results_from_paths(
                    compressed_path=compressed_path,
                    reconstructed_path=shuffled_path)
            decompression_results.reconstructed_path = reconstructed_path

            # Unshuffle and write output
            array = enb.isets.load_array(shuffled_path)
            time_before = time.process_time() if not enb.config.options.report_wall_time else time.time()
            unshuffled_array = bitshuffle.bitunshuffle(np.ascontiguousarray(array))
            time_after = time.process_time() if not enb.config.options.report_wall_time else time.time()
            shuffle_time = max(0.0, time_after - time_before)
            enb.isets.dump_array(array=unshuffled_array, file_or_path=reconstructed_path)
            os.remove(shuffled_path)

            if decompression_results.decompression_time_seconds is None:
                decompression_results.decompression_time_seconds = decompression_time
            decompression_results.decompression_time_seconds += shuffle_time
            
            return decompression_results

    @property
    def name(self):
        """Return the original codec name and the quantization parameter
        """
        return f"{self.codec.name}_bitshuffle"

    @property
    def label(self):
        """Return the original codec label and the quantization parameter.
        """
        return f"Bitshuffle + {self.codec.label}"



import textwrap
import enb.plugins

class FPackPlugin(enb.plugins.PluginMake):
    name = "fpack"
    label = "Wrapper for the FPACK compressor"
    tags = {"data compression", "image", "codec"}
    contrib_authors = ["William D. Pence"]
    contrib_reference_urls = ["https://heasarc.gsfc.nasa.gov/fitsio/fpack/"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/"
         "cfitsio-3.49.tar.gz?raw=true",
         "cfitsio-3.49.tar.gz")]
    extra_requirements_message = textwrap.dedent("""
        The cmake tool is needed to build this plugin, if not present already. 
        You can install cmake as follows: 

        * Ubuntu: `sudo apt install cmake`
        * MacOS:
            1. Install homebrew from http://brew.sh
            2. `brew install gcc@9`
        """)
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Codec wrapper for the SPECK codec.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2022/04/08"

import os

import enb


class SPECK(enb.icompression.LosslessCodec,
            enb.icompression.LossyCodec,
            enb.icompression.JavaWrapperCodec,
            enb.icompression.GiciLibHelper):
    """Wrapper for the SPECK codec.
    """

    def __init__(self,
                 compressor_jar=os.path.join(os.path.dirname(__file__), "SpeckCode.jar"),
                 decompressor_jar=os.path.join(os.path.dirname(__file__), "SpeckDecode.jar"),
                 wavelet_type="1", wavelet_levels=5):
        """
        :param wavelet_type:     Discrete wavelet transform type for each image component.
            First value is for the first component, second value for the second component and so on.
            If only one value is specified, wavelet transform type will be the same for all components.
            Valid values are:
                 0 -  No wavelet transform
                 1 - Integer (Reversible) 5/3 DWT
                 2 - Real Isorange (irreversible) 9/7 DWT
                 3 - Real Isonorm (irreversible) 9/7 DWT
                 4 - Integer (Reversible) 9/7M DWT (CCSDS-Recommended)
                 5 - Integer 5/3 DWT (classic construction)
                 6 - Integer 9/7 DWT (classic construction)
                 7 - Haar Transform
                 8 - Reversible-Haar Transform
        :param wavelet_levels: number of wavelet levels to be applied.
        """
        super().__init__(compressor_jar=compressor_jar,
                         decompressor_jar=decompressor_jar,
                         param_dict=dict(wavelet_type=wavelet_type,
                                         wavelet_levels=wavelet_levels))

    @property
    def label(self):
        return "SPECK"

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        assert original_file_info["component_count"] == 1, \
            f"Only images with 1 component are currently supported by {self.__class__.__name__}"
        assert original_file_info["big_endian"], \
            f"Only big-endian samples are currently supported by {self.__class__.__name__}"
        assert original_file_info["bytes_per_sample"] in [1, 2], \
            f"Only 8-bit and 16-bit samples are currently supported by {self.__class__.__name__}"
        
        return f"-Xmx256g -jar {self.compressor_jar} -i {original_path} -f {compressed_path} " \
               f"-g {self.get_gici_geometry_str(original_file_info=original_file_info)} " \
               f"-w {self.param_dict['wavelet_type']} -wl {self.param_dict['wavelet_levels']}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        assert original_file_info["component_count"] == 1, f"Only images with 1 component are currently supported."
        
        return f"-Xmx256g -jar {self.decompressor_jar} " \
               f"-f {compressed_path} -o {reconstructed_path} " \
               f"-g {self.get_gici_geometry_str(original_file_info=original_file_info)} " \
               f"-w {self.param_dict['wavelet_type']} -wl {self.param_dict['wavelet_levels']}"



#!/usr/bin/env python3
"""Entry point for the main enb CLI.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/08/01"

import os
import sys
import argparse
import textwrap
import enb
from enb.config import options


def _get_cli_parser():
    """Produce and return the main argument parser for enb.
    """
    cli_parser = argparse.ArgumentParser(
        prog="enb",
        description="CLI to the experiment notebook (enb) framework (see "
                    "https://github.com/miguelinux314/experiment-notebook)."
                    "\n\nSeveral subcommands are available; "
                    "use `enb <subcommand> -h` to show help about "
                    "any specific command.",
        formatter_class=argparse.RawTextHelpFormatter)
    cli_parser.subparsers = cli_parser.add_subparsers(
        dest="command", required=True,
        description="Available enb CLI commands.")

    # plugin subcommand
    cli_parser.plugin_parser = cli_parser.subparsers.add_parser(
        "plugin", help="Install and manage plugins.")
    cli_parser.plugin_parser.subparsers = cli_parser.plugin_parser.add_subparsers(
        description="Plugin subcommands", dest="subcommand", required=True)
    # # plugin install
    cli_parser.plugin_parser.install_parser = cli_parser.plugin_parser.subparsers.add_parser(
        "install", help="Install an available plugin.")
    cli_parser.plugin_parser.install_parser.add_argument(
        "plugin_name",
        help="Name of the plugin to be installed.")
    cli_parser.plugin_parser.install_parser.add_argument(
        "destination_dir",
        help="Path to the directory that will contain the installed plugin. "
             "Defaults to the working dir.",
        nargs="?",
        default=None)
    cli_parser.plugin_parser.install_parser.add_argument(
        # Used to trigger the desired call and save the return status
        nargs=0, dest="", action=PluginInstall)

    # # plugin list
    cli_parser.plugin_parser.list_parser = \
        cli_parser.plugin_parser.subparsers.add_parser(
            "list", help="List available plugins.")
    cli_parser.plugin_parser.list_parser.add_argument(
        "-v", action="count", default=0,
        help="Show additional information about the available plugins.")
    filtering_group = cli_parser.plugin_parser.list_parser.add_argument_group(
        "Filtering options")
    filtering_group.add_argument(
        "--exclude", nargs="*", metavar="exclude_name", default=[], type=str,
        required=False,
        help="If provided, plugins matching any of these arguments are not listed. "
             "It overwrites the filter argument(s).")
    filtering_group.add_argument(
        "filter", nargs="*",
        help="If provided, only plugins matching passed string(s) are listed.")

    cli_parser.plugin_parser.list_parser.add_argument(
        # Used to trigger the desired call and save the return status
        nargs=0, dest="", action=PluginList)

    # # show subcommand
    cli_parser.show_parser = cli_parser.subparsers.add_parser(
        "show", help="Show useful information about enb and enb projects.")
    cli_parser.show_parser.subparsers = cli_parser.show_parser.add_subparsers(
        description="Show subcommands", dest="subcommand", required=True)
    ## show styles
    cli_parser.show_parser.styles_parser = \
        cli_parser.show_parser.subparsers.add_parser(
            "styles", help="Show available style names for plotting.")
    cli_parser.show_parser.styles_parser.add_argument(
        "filter", nargs="*", help="Show only styles containing this string.")
    cli_parser.show_parser.styles_parser.add_argument(
        nargs=0, dest="", action=ShowStyles)

    # Help command
    cli_parser.help_parser = cli_parser.subparsers.add_parser(
        "help", help="Show this help and exit.")

    return cli_parser


class PluginInstall(argparse.Action):
    """Action for installing an installable (Plugin, Template, etc).
    """

    def __call__(self, parser, namespace, values, option_string=None):
        # Get plugin name and instance
        plugin_name = namespace.plugin_name.strip()
        assert " " not in plugin_name, \
            f"Plugin names cannot have spaces: {repr(plugin_name)}"
        try:
            plugin = [p for p in enb.plugins.list_all_installables() if
                      p.name == plugin_name][0]
        except IndexError:
            enb.logger.error(
                f"Invalid plugin name {repr(plugin_name)}. "
                "Run `enb plugin list` to see all available plugins, "
                "or `enb plugin list <something>` to filter that list.")
            sys.exit(1)

        # Configure output paths
        if namespace.destination_dir is None and not os.path.exists("plugins"):
            os.makedirs("plugins")
            with open(os.path.join("plugins", "__init__.py"), "w"):
                pass
        destination_dir = namespace.destination_dir or os.path.join("plugins", plugin_name)
        if os.path.exists(destination_dir) and not issubclass(plugin, enb.plugins.Template):
            enb.logger.error(f"Error installing {repr(plugin.name)}.\n"
                             f"The destination dir {repr(destination_dir)} "
                             "already exists. Remove and try again.")
            sys.exit(1)
            
        # Install
        try:
            plugin.install(installation_dir=destination_dir)
        except (SyntaxError, ValueError) as ex:
            enb.logger.error(
                f"Error installing plugin {repr(plugin_name)} into {repr(os.path.abspath(destination_dir))}. "
                f"The following exception was raised:\n{ex}")
            sys.exit(1)

        # Set status
        setattr(namespace, self.dest, 0)


class PluginList(argparse.Action):
    """Action for listing available plugins.
    """

    def installable_matches_querys(self, installable, query_list):
        """Return true if and only if the installable matches any of the
        provided queries.
        """
        # pylint: disable=no-self-use
        return any(query.lower() in installable.name.lower() or
                   (installable.label.lower()
                    and query in installable.label.lower()) or
                   any(query.lower() in author.lower() for author in
                       installable.contrib_authors)
                   or any(query.lower() in t
                          for t in installable.tags)
                   or any(query.lower() in t
                          for t in installable.tested_on)
                   for query in query_list)

    def __call__(self, parser, namespace, values, option_string=None):
        # Process input and obtain the list of installables (plugins)
        # that match the query (all if no query was presented)
        all_installables = enb.plugins.list_all_installables()
        filtered_installables = [
            i for i in all_installables
            if self.installable_matches_querys(
                i, namespace.filter if namespace.filter else [])] \
            if namespace.filter else all_installables
        if namespace.exclude:
            filtered_installables = [
                i for i in filtered_installables
                if not self.installable_matches_querys(
                    installable=i, query_list=namespace.exclude)]

        # Print mathing plugins
        self.print_matching_plugins(all_installables=all_installables,
                                    filtered_installables=filtered_installables,
                                    namespace=namespace)
        print()

        # Print available tags
        self.print_available_tags()

        # Print extra info
        if not options.verbose and "-h" not in sys.argv:
            print()
            print("Run" + \
                  (' with -v for authorship and additional information,'
                   if not enb.config.options.verbose else '') + \
                  " with -h for full help.")

    def print_available_tags(self):
        """Show information about what tags that have been defined and
        # can be used for filtering.
        """
        # pylint: disable=no-self-use
        print("The following plugin tags have been defined and can be "
              "used for filtering:\n")
        max_tag_length = max(
            len(t) for t in
            enb.plugins.installable.InstallableMeta.tag_to_installable.keys())
        tag_fmt_str = f"{{tag:{max_tag_length}s}}"
        for tag, installable_list in sorted(
                enb.plugins.installable.InstallableMeta.tag_to_installable.items(),
                key=lambda t: list(
                    enb.plugins.installable.tag_to_description.keys()).index(
                    t[0])):
            print(
                f"  - {tag_fmt_str.format(tag=tag)} ({len(installable_list):3d} "
                f"{'' if len(installable_list) != 1 else ' '}"
                f"plugin{'s' if len(installable_list) != 1 else ''})",
                end="")
            try:
                print(f" {enb.plugins.installable.tag_to_description[tag]}",
                      end="")
                print("" if enb.plugins.installable.tag_to_description[
                    tag].endswith(".") else ".")
            except KeyError:
                print()

    def print_matching_plugins(
            self, all_installables, filtered_installables, namespace):
        """Print the list of matching plugins.
        """
        # pylint: disable=no-self-use
        if namespace.filter and not filtered_installables:
            print(f"No plugin matched the filter criteria "
                  f"({', '.join(repr(f) for f in namespace.filter)}).")
        else:
            print(f"Showing {len(filtered_installables)} plugins", end="")
            if namespace.filter:
                print(
                    f" matching {'any of ' if len(namespace.filter) > 1 else ''}"
                    f"{', '.join(repr(f) for f in namespace.filter)}, "
                    f"out of {len(all_installables)} available)", end="")
            else:
                print(".\nYou can add arguments to filter this list, "
                      "and/or use the --exclude argument.\n"
                      "Add -v for extra information on the listed plugins",
                      end="")
            print(".\n")

            for installable in filtered_installables:
                label = installable.label if installable.label else ''
                label = f"{label} (privative)" if "privative" in installable.tags else label
                while label[-1] == ".":
                    label = label[:-1]
                while "  " in label:
                    label = label.replace("  ", "")
                label = label.strip()
                if label:
                    label = label[0].upper() + label[1:]
                if options.verbose:
                    print("-" * 20 + f"  {installable.name} :: ", end="")
                else:
                    print(f"{installable.name:>25s} :: ", end="")
                print("\n".join(textwrap.wrap(label, 100)), end="")
                print(".")
                if enb.config.options.verbose:
                    installable.print_info()


class ShowStyles(argparse.Action):
    """Show the list of available styles for plotting.
    """

    def __call__(self, parser, namespace, values, option_string=None):
        print("The following styles are available for plotting:\n\n\t- ",
              end="")
        print("\n\t- ".join(
            repr(s) for s in enb.plotdata.get_available_styles()))


def main():
    """Entry point for the enb CLI (not just importing enb from a script).
    """
    cli_parser = _get_cli_parser()
    cli_options, _ = cli_parser.parse_known_args()

    if cli_options.command is None:
        print("No command provided. Showing help instead.\n")
        print(f"{' [ enb help ] ':-^80s}")
        cli_parser.print_help()
    elif cli_options.command == "help":
        cli_parser.print_help()
        return
    else:
        # Successful command run
        enb.logger.verbose("")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""Wrappers for E. Maglis's M-CALIC implementation
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/30"

import os
import sortedcontainers
import tempfile
import numpy as np
import glob
import copy
import subprocess
import shutil
import random

from enb import icompression
from enb import isets
from enb import tarlite
from enb.config import options


class MCALIC_Magli(icompression.LosslessCodec, icompression.NearLosslessCodec, icompression.WrapperCodec):
    FORMAT_BSQ, FORMAT_BIL = range(2)
    default_format = FORMAT_BSQ

    max_tested_spatial_size = 17418240
    max_dimension_size = 2500
    split_height_count = 3
    split_width_count = 3

    def __init__(self, max_error=0, bin_dir=None, data_format=None, output_invocation_dir=None):
        """
        :param max_error: maximum pixelwise error allowed. Use 0 for lossless
          compression
        :param bin_dir: path to the directory that contains the
          ldc_encoder, ldc_decoder and ldc_header_tool binaries. If it is None,
          options.external_bin_base_dir is None. If this is None as well, the
          same directory of this script is used by default.
        :param data_format: bsq/bil format of the expected data.
          If none, the default (BSQ) is used
        """
        bin_dir = bin_dir if bin_dir is not None else options.external_bin_base_dir
        bin_dir = bin_dir if bin_dir is not None else os.path.dirname(__file__)
        assert os.path.isdir(bin_dir), f"Invalid binary dir {bin_dir}."

        param_dict = sortedcontainers.SortedDict()
        max_error = int(max_error)
        assert max_error >= 0, f"Invalid max_error {max_error}"
        param_dict["max_error"] = max_error
        data_format = data_format if data_format is not None else self.default_format
        assert data_format in [self.FORMAT_BSQ, self.FORMAT_BIL], f"Invalid data format {data_format}"
        param_dict["data_format"] = data_format
        icompression.WrapperCodec.__init__(
            self, compressor_path=os.path.join(bin_dir, "Mcalic_enc_nl"),
            decompressor_path=os.path.join(bin_dir, "Mcalic_dec_nl"),
            param_dict=param_dict, output_invocation_dir=output_invocation_dir)

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        # Tested limit: self.max_tested_spatial_size
        if original_file_info["width"] <= self.max_dimension_size \
                and original_file_info["height"] <= self.max_dimension_size:
            return self.compress_one(original_path=original_path,
                                     compressed_path=compressed_path,
                                     original_file_info=original_file_info)
        else:
            tl_writer = tarlite.TarliteWriter()
            img = isets.load_array_bsq(file_or_path=original_path, image_properties_row=original_file_info)
            with tempfile.TemporaryDirectory(dir=options.base_tmp_dir) as tmp_dir:
                compound_size = 0
                total_compression_time = 0
                for y in range(self.split_height_count):
                    for x in range(self.split_width_count):
                        small_array = \
                            img[x * (original_file_info["width"] // self.split_width_count):
                                (((x + 1) * (original_file_info[
                                                 "width"] // self.split_width_count)) if x < self.split_width_count - 1 else
                                 original_file_info["width"]),
                            y * (original_file_info["height"] // self.split_height_count):
                            (((y + 1) * (original_file_info[
                                             "height"] // self.split_height_count)) if y < self.split_height_count - 1 else
                             original_file_info["height"]),
                            :]
                        small_path = os.path.join(tmp_dir, f"{x}_{y}.raw")
                        small_compressed_path = os.path.join(tmp_dir, f"{x}_{y}.mcalic")
                        isets.dump_array_bsq(small_array, small_path)
                        small_file_info = copy.copy(original_file_info)
                        small_file_info["width"], small_file_info["height"], small_file_info[
                            "component_count"] = small_array.shape
                        compression_results = self.compress_one(
                            original_path=small_path, compressed_path=small_compressed_path,
                            original_file_info=small_file_info)
                        total_compression_time += compression_results.compression_time_seconds
                        tl_writer.add_file(small_compressed_path)
                        os.remove(small_path)
                        compound_size += small_array.size

                assert compound_size == original_file_info[
                    "samples"], f"compound_size = {compound_size} != {original_file_info['samples']} = original samples"
                tl_writer.write(output_path=compressed_path)

                compression_results = self.compression_results_from_paths(original_path=original_path,
                                                                          compressed_path=compressed_path)
                compression_results.compression_time_seconds = total_compression_time
                return compression_results

    def compress_one(self, original_path: str, compressed_path: str, original_file_info=None):
        """Compress one image tile with M-CALIC.
        """
        assert original_file_info["bytes_per_sample"] == 2, \
            f"This implementation of M-CALIC ({self.compressor_path}) only supports 16bpp"
        assert original_file_info["component_count"] > 1, \
            f"This implementation of M-CALIC ({self.compressor_path}) only supports images with more than one component"


        with tempfile.NamedTemporaryFile(
                dir=options.base_tmp_dir, prefix=f"bil_le_{os.path.basename(original_path)}") as bil_le_file:
            # M-Calic implementation requires little endian, unsigned 16bpp BIL format
            original_dtype = isets.iproperties_row_to_numpy_dtype(image_properties_row=original_file_info)
            img = np.fromfile(original_path, dtype=original_dtype).reshape(
                original_file_info["component_count"], original_file_info["height"], original_file_info["width"])
            

            offset = None
            if original_file_info["signed"]:
                offset, original_max = int(img.min()), int(img.max())
                offset = min(offset, 0)
                assert original_max - offset <= 2 ** 15 - 1, \
                    f"Invalid dynamic range of signed image ({offset}, {original_max})"
                img = (img.astype("i4") - offset).astype(original_dtype.replace("i", "u"))

            if original_file_info["big_endian"]:
                img = img.astype(original_dtype.replace(">", "<").replace("i", "u"))
            img.swapaxes(0, 1).tofile(bil_le_file.name)

            if original_file_info["signed"]:
                with tempfile.NamedTemporaryFile(dir=options.base_tmp_dir,
                                                 prefix=f"bil_le_{os.path.basename(original_path)}",
                                                 suffix=".mcalic") as tmp_compressed_file, \
                        tempfile.NamedTemporaryFile(dir=options.base_tmp_dir,
                                                    prefix=f"side_info_{os.path.basename(original_path)}",
                                                    suffix=".txt", mode="w") as si_file:
                    si_file.write(f"{abs(offset):d}")
                    si_file.flush()
                    compression_results = super().compress(original_path=bil_le_file.name,
                                                           compressed_path=tmp_compressed_file.name,
                                                           original_file_info=original_file_info)
                    tarlite.TarliteWriter(initial_input_paths=[si_file.name, tmp_compressed_file.name]).write(
                        compressed_path)
                    compression_results.original_path = original_path
                    compression_results.compressed_path = compressed_path
                    return compression_results
            else:
                compression_results = super().compress(
                    original_path=bil_le_file.name, compressed_path=compressed_path,
                    original_file_info=original_file_info)
                compression_results.original_path = original_path
                return compression_results

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        s = f"{original_path} {compressed_path} " \
            f"{original_file_info['component_count']} {original_file_info['height']} {original_file_info['width']} " \
            f"{8 * original_file_info['bytes_per_sample']} {self.param_dict['max_error']} {self.param_dict['data_format']}"
        return s

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        if original_file_info["width"] <= self.max_dimension_size and original_file_info[
            "height"] <= self.max_dimension_size:
            return self.decompress_one(compressed_path=compressed_path,
                                       reconstructed_path=reconstructed_path,
                                       original_file_info=original_file_info)
        else:
            tl_reader = tarlite.TarliteReader(tarlite_path=compressed_path)
            img = np.zeros(
                (original_file_info["width"], original_file_info["height"], original_file_info["component_count"]),
                dtype=isets.iproperties_row_to_numpy_dtype(image_properties_row=original_file_info))
            total_decompression_time = 0
            with tempfile.TemporaryDirectory(dir=options.base_tmp_dir) as tmp_dir:
                tl_reader.extract_all(output_dir_path=tmp_dir)
                invocation = f"ls -lah {tmp_dir}"
                status, output = subprocess.getstatusoutput(invocation)
                if status != 0:
                    raise Exception(f"Status = {status} != 0.\n"
                                    f"Input=[{invocation}].\n"
                                    f"Output=[{output}]")

                for y in range(self.split_height_count):
                    for x in range(self.split_width_count):
                        small_compressed_path = os.path.join(tmp_dir, f"{x}_{y}.mcalic")
                        assert os.path.exists(small_compressed_path)
                        small_path = os.path.join(tmp_dir, f"{x}_{y}.raw")
                        small_file_info = copy.copy(original_file_info)

                        small_file_info["height"] = original_file_info["height"] // self.split_height_count \
                            if y < self.split_height_count - 1 \
                            else original_file_info["height"] - (self.split_height_count - 1) * (
                                original_file_info["height"] // self.split_height_count)

                        small_file_info["width"] = original_file_info["width"] // self.split_width_count \
                            if x < self.split_width_count - 1 \
                            else original_file_info["width"] - (self.split_width_count - 1) * (
                                original_file_info["width"] // self.split_width_count)

                        dr = self.decompress_one(compressed_path=small_compressed_path, reconstructed_path=small_path,
                                                 original_file_info=small_file_info)
                        total_decompression_time += dr.decompression_time_seconds
                        small_array = isets.load_array_bsq(file_or_path=small_path,
                                                           image_properties_row=small_file_info)
                        img[x * (original_file_info["width"] // self.split_width_count):
                            (((x + 1) * (original_file_info[
                                             "width"] // self.split_width_count)) if x < self.split_width_count - 1 else
                             original_file_info["width"]),
                        y * (original_file_info["height"] // self.split_height_count):
                        (((y + 1) * (original_file_info[
                                         "height"] // self.split_height_count)) if y < self.split_height_count - 1 else
                         original_file_info["height"]),
                        :] = small_array
                isets.dump_array_bsq(array=img, file_or_path=reconstructed_path)

            decompression_results = self.decompression_results_from_paths(
                compressed_path=compressed_path, reconstructed_path=reconstructed_path)
            decompression_results.decompression_time_seconds = total_decompression_time
            return decompression_results

    def decompress_one(self, compressed_path, reconstructed_path, original_file_info=None):
        total_decompression_time = 0
        with tempfile.NamedTemporaryFile(
                dir=options.base_tmp_dir,
                prefix=f"reconstructed_{os.path.basename(reconstructed_path)}",
                suffix=".raw") as bil_le_file:
            offset = 0
            if original_file_info["signed"]:
                reader = tarlite.TarliteReader(compressed_path)
                with tempfile.TemporaryDirectory(dir=options.base_tmp_dir) as tmp_extract_dir:
                    reader.extract_all(tmp_extract_dir)

                    import subprocess
                    invocation = f"ls -lah {tmp_extract_dir}"
                    status, output = subprocess.getstatusoutput(invocation)
                    if status != 0:
                        raise Exception(f"Status = {status} != 0.\n"
                                        f"Input=[{invocation}].\n"
                                        f"Output=[{output}]")

                    with open(glob.glob(os.path.join(tmp_extract_dir, "*.txt"))[0]) as si_file:
                        offset = int(si_file.read())
                        assert offset >= 0

                    os.path.getsize(tmp_extract_dir)
                    inner_compressed_path = glob.glob(os.path.join(tmp_extract_dir, "*.mcalic"))[0]

                    dr = self.decompress_short_names(compressed_path=inner_compressed_path,
                                                     reconstructed_path=bil_le_file.name,
                                                     original_file_info=original_file_info)
                    total_decompression_time += dr.decompression_time_seconds
            else:
                dr = self.decompress_short_names(compressed_path=compressed_path, reconstructed_path=bil_le_file.name,
                                                 original_file_info=original_file_info)
                total_decompression_time += dr.decompression_time_seconds

            original_dtype = isets.iproperties_row_to_numpy_dtype(image_properties_row=original_file_info)

            img = np.fromfile(bil_le_file.name, dtype=original_dtype.replace(">", "<").replace("i", "u")).reshape(
                original_file_info["height"], original_file_info["component_count"], original_file_info["width"])

            if original_file_info["signed"]:
                if offset != 0:
                    img = (img.astype("i4") - offset).astype(original_dtype)
                else:
                    img = img.astype(original_dtype)

            img = img.swapaxes(0, 1)
            if original_file_info["big_endian"] and not original_file_info["signed"]:
                # Signed file are already converted back to big_endian if necessary in the previous call to astype()
                img = img.astype(original_dtype)
            img.tofile(reconstructed_path)

        try:
            # The decoder always produces this file
            os.remove("seqRec")
        except FileNotFoundError:
            pass

        decompression_results = self.decompression_results_from_paths(compressed_path=compressed_path,
                                                                      reconstructed_path=reconstructed_path)
        decompression_results.decompression_time_seconds = total_decompression_time
        return decompression_results

    def decompress_short_names(self, compressed_path, reconstructed_path, original_file_info):
        """Binary seems to have problems with too long file names
        """
        c_path = f"{random.randint(0, 10000000)}.raw"
        r_path = f"{random.randint(0, 10000000)}.raw"
        try:
            shutil.copyfile(compressed_path, c_path)
            decompression_results = super().decompress(compressed_path=c_path,
                                                       reconstructed_path=r_path,
                                                       original_file_info=original_file_info)
            shutil.copyfile(r_path, reconstructed_path)
            decompression_results.compressed_path = compressed_path
            decompression_results.reconstructed_path = reconstructed_path
            return decompression_results
        finally:
            for p in (c_path, r_path):
                if os.path.exists(p):
                    os.remove(p)

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"{reconstructed_path} {compressed_path} " \
               f"{original_file_info['component_count']} {original_file_info['height']} {original_file_info['width']} " \
               f"{8*original_file_info['bytes_per_sample']} {self.param_dict['max_error']} {self.param_dict['data_format']}"

    @property
    def label(self):
        return "M-CALIC"



import enb.plugins


class HDF5Plugin(enb.plugins.Plugin):
    name = "hdf5"
    label = "Codec wrappers for the h5py library"
    tags = {"data compression", "image", "codec"}
    contrib_authors = ["Andrew Collete et al."]
    contrib_reference_urls = ["https://github.com/h5py/h5py"]
    required_pip_modules = ["h5py"]
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Example that approximates the value of pi using the montecarlo method.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/12/27"

import platform

import numpy as np
import enb


class MontecarloPiExperiment(enb.experiment.Experiment):
    def column_sample_size(self, index, row):
        """Keep track of how many samples were employed for this approximation.
        """
        task = self.index_to_path_task(index)[1]
        return task.param_dict["sample_size"]

    def column_computed_on_ip(self, index, row):
        """Register the node that computed this particular sample.
        """
        return enb.misc.get_node_ip()

    def column_estimated_pi_value(self, index, row):
        """Make a Montecarlo approximation of pi.
        """
        random_values = -1 + 2 * np.random.random((row["sample_size"], 2))
        return 4 * np.count_nonzero(
            random_values[:, 0] * random_values[:, 0] + random_values[:, 1] * random_values[:, 1]
            <= 1) / row["sample_size"]


if __name__ == '__main__':
    sample_size = 10000000
    sample_count = 128

    if enb.config.options.ssh_cluster_csv_path is None:
        if enb.parallel_ray.is_ray_enabled():
            print("enb.config.options.ssh_cluster_csv_path is not set. "
                  "You can do so with --ssh_cluster_csv_path=enb_cluster.csv"
                  "or creating an '*.ini' file with\n"
                  "[enb.config.options]\nssh_cluster_csv_path = enb_cluster.csv\n"
                  "Please see https://miguelinux314.github.io/experiment-notebook/cluster_setup.html "
                  "for more details.")
    else:
        print(f"Using cluster configuration file at {enb.config.options.ssh_cluster_csv_path}")

    exp = MontecarloPiExperiment(
        tasks=[enb.experiment.ExperimentTask(dict(label=i, sample_size=sample_size))
               for i in range(sample_count)])
    df = exp.get_df()

    total_samples = df["sample_size"].sum()
    estimated_pi_value = (df["estimated_pi_value"] * df["sample_size"]).sum() / total_samples
    different_nodes = len(df["computed_on_ip"].unique())
    print(f"Approximated pi ~ {estimated_pi_value} using {total_samples} total samples on {different_nodes} nodes.")


#!/usr/bin/env python3
"""Miscellaneous tools for `enb`.

This module does not and should not import anything from enb, so that other
modules may use misc tools at definition time."""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/07/11"

import re
import pdb
import signal
import socket
import csv
import time
import rich


def get_defining_class_name(method):
    """Return the name of the class of which f is a method, or None if not
    bound to any class.
    """
    try:
        return method.__qualname__.split('.<locals>', 1)[0].rsplit('.')[-2]
    except IndexError:
        return None


def remove_argparse_action(parser, action):
    """Entirely remove an action from a parser, from its subparsers and
    groups if it exists. Adapted from https://stackoverflow.com/a/49753634.
    """
    # pylint: disable=protected-access,too-many-branches
    try:
        parser._remove_action(action)
    except ValueError:
        pass
    try:
        parser._actions.remove(action)
    except ValueError:
        pass

    for group in parser._action_groups:
        try:
            group._remove_action(action)
        except ValueError:
            pass
        try:
            group._actions.remove(action)
        except ValueError:
            pass
        try:
            group._group_actions.remove(action)
        except ValueError:
            pass

    old_action = action
    try:
        parser._remove_action(old_action)
    except ValueError:
        pass
    try:
        parser._actions.remove(old_action)
    except ValueError:
        pass

    for option_str in old_action.option_strings:
        try:
            del parser._option_string_actions[option_str]
        except KeyError:
            pass

    vars_action = vars(old_action)
    try:
        var_group_actions = vars_action['_group_actions']
    except KeyError:
        var_group_actions = None
    if var_group_actions is not None:
        for group_action in var_group_actions:
            if group_action.dest == action:
                var_group_actions.remove(group_action)


def split_camel_case(camel_string):
    """Split a camel case string like ThisIsAClass into a string like "This
    Is A Class".
    """
    return " ".join(
        re.findall(r"[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))", camel_string))


def get_all_subclasses(*base_classes):
    """Return a set of all subclasses of the classes in base_classes,
    which have been defined at this point.

    The base classes are never returned as subclasses.

    :param base_classes: the list of classes for which subclasses are to be found
    """

    def get_subclasses_recursive(cls):
        direct_subclasses = set(cls.__subclasses__())
        recursive_subclasses = set()
        for subclass in direct_subclasses:
            recursive_subclasses = recursive_subclasses.union(
                get_subclasses_recursive(subclass))
        return direct_subclasses.union(recursive_subclasses)

    base_classes = set(base_classes)
    all_subclasses = set()

    for base_class in base_classes:
        all_subclasses = all_subclasses.union(
            get_subclasses_recursive(base_class))

    return set(cls for cls in all_subclasses if cls not in base_classes)


class Singleton(type):
    """Classes using this as metaclass will only be instantiated once.
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        """This method replaces the regular initializer of classes with this
        as their metaclass. `*args` and `**kwargs` are passed directly to
        their initializer and do not otherwise affect the Singleton behavior.
        """
        try:
            return cls._instances[cls]
        except KeyError:
            cls._instances[cls] = super().__call__(*args, **kwargs)
            return cls._instances[cls]


class ExposedProperty:
    """This method can be used to expose object properties as public callables
    that return what requesting that property would.
    """

    # pylint: disable=too-few-public-methods
    def __init__(self, instance, property_name):
        self.property_name = property_name
        self.instance = instance

    def __call__(self, *args, **kwargs):
        return getattr(self.instance, self.property_name)


class CircularList(list):
    """A tuned list that automatically applies modulo len(self) to the given
    index, allowing for circular, index-based access to the data (whereas
    itertools.cycle does not allow accessing elements by index).
    """

    def __getitem__(self, item):
        return super().__getitem__(item % len(self))


class LapTimer:
    """Keep track of time duration similar to a lap timer. Useful to track
    the time elapsed between consecutive calls to print_lap.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self):
        self.last_time = time.time()

    def print_lap(self, msg=None):
        """Print the elapsed time since the last time this method was called,
        or when this instance was created if it is the first time this method
        is called.
        """
        print(f"Elapsed{' ' + msg if msg is not None else ''}: "
              f"{time.time() - self.last_time}")
        self.last_time = time.time()


def class_to_fqn(cls):
    """Given a class (type instance), return its fully qualified name (FQN).
    """
    return f"{str(cls.__module__) + '.' if cls.__module__ is not None else ''}" \
           f"{cls.__name__}"


def csv_to_latex_tabular(input_csv_path, output_tex_path, contains_header=True,
                         use_booktabks=True):
    """Read a CSV table from a file and output it as a latex table to another
    file. The first row is assumed to be the header.

    :param input_csv_path: path to a file containing CSV data.
    :param output_tex_file: path where the tex contents are to be stored,
      ready to be added to latex with the `input` command.
    :param contains_header: if True, the first line is assumed to be a header
      containing column names.
    :param use_booktabs: if True, a booktabs-based decoration style is used
      for the table. Otherwise, standard latex is used only.
    """
    with open(input_csv_path, "r", encoding="utf-8") as csv_file, \
            open(output_tex_path, "w", encoding="utf-8") as tex_file:
        tex_file.write("\\begin{tabular}{")

        for i, row in enumerate(csv.reader(csv_file)):
            if i == 0:
                tex_file.write("l" * len(row) + "}\n")

            if i == 0 and contains_header:
                tex_file.write("\\toprule\n" if use_booktabks else "\\hline\n")
                tex_file.write(
                    " & ".join(f"\\textbf{{{escape_latex(c)}}}" for c in row) + r" \\" + "\n")
                tex_file.write("\\midrule\n" if use_booktabks else "\\hline\n")
            else:
                tex_file.write(" & ".join(escape_latex(row)) + r" \\" + "\n")

        tex_file.write("\\bottomrule\n" if use_booktabks else "\\hline\n")
        tex_file.write("\\end{tabular}\n")

def escape_latex(s):
    """Return a latex-scaped version of string s.
    """
    return s.replace("\\", "\\\\").replace("_", r"\_").replace(r"%", r"\%").replace(r"&", r"\&")


def get_node_ip():
    """Get the current IP address of this node.
    """
    soc = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    soc.connect(("8.8.8.8", 80))
    address = soc.getsockname()[0]
    soc.close()
    return address


def get_node_name():
    """Get the host name of this node. Alias for `socket.gethostname`.
    """
    return socket.gethostname()


def capture_usr1():
    """Capture the reception of a USR1 signal into pdb.

    From http://blog.devork.be/2009/07/how-to-bring-running-python-program.html.
    """

    def handle_pdb(sig, frame):
        # pylint: disable=unused-argument
        print("\n" * 2)
        print("Captured USR1 signal! Activating pdb...")
        print("\n" * 2)
        pdb.Pdb().set_trace(frame)

    try:
        signal.signal(signal.SIGUSR1, handle_pdb)
    except AttributeError:
        # Only available in unix platforms 
        pass

class BootstrapLogger:
    """Imitate enb.log.Logger's interface before it is loaded. This is needed to solve circular imports,
    i.e., when an error with the managed attributes decorator takes place before the full logger is available
    (within the config submodule).
    """

    def __init__(self):
        from .config.aoptions import get_options
        self.options = get_options()
        self.console = rich.console.Console(highlight=False, markup=False)

    def log(self, *args, style=None, **kwargs):
        self.console.print(*args, **kwargs, style=style, highlight=False, markup=False)

    def core(self, *args, **kwargs):
        self.log(*args, **kwargs, style="#28c9ff on #000000")

    def error(self, *args, **kwargs):
        self.log(*args, **kwargs, style="#ff5255 on #000000")

    def warn(self, *args, **kwargs):
        self.log(*args, **kwargs, style="#ffca4f on #000000")

    def message(self, *args, **kwargs):
        self.log(*args, **kwargs, style="#28c9ff on #000000")

    def verbose(self, *args, **kwargs):
        if self.options.verbose >= 1:
            self.log(*args, **kwargs, style="#c8ffc8 on #000000")

    def info(self, *args, **kwargs):
        if self.options.verbose >= 2:
            self.log(*args, **kwargs, style="#afffbe on #000000")

    def debug(self, *args, **kwargs):
        if self.options.verbose >= 3:
            self.log(*args, **kwargs, style="#909090 on #000000")



import os
import enb.plugins


class IRAFPhotometryPlugin(enb.plugins.Plugin):
    name = "iraf_photometry"
    label = "Extra photometry information from image files"
    tags = {"image"}
    contrib_authors = ["IRAF community"]
    contrib_reference_urls = ["https://iraf-community.github.io"]
    required_pip_modules = ["pyraf"]
    tested_on = {"linux"}
    extra_requirements_message = """
        The IRAF photometry plugin requires the `iraf` package and `pyraf` pip module to be installed in your system. 
        In ubuntu/debian/etc it can be installed with `sudo apt install iraf`. 
        It might be available in other distribution repositories as well. 
        You can also install IRAF from the sources available at https://github.com/iraf-community/iraf.
        The pyraf python module can also be installed from pip, but it requires non-automatic configuration.   
        """



#!/usr/bin/env python3
"""Wrappers for the reference JPEG implementation
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/08"

import os
import tempfile
import numpy as np
import sortedcontainers

import enb
from enb import icompression
from enb import isets
from enb import pgm
from enb import tarlite
from enb.config import options


class Abstract_JPEG(icompression.WrapperCodec):
    max_dimension_size = 65535

    def __init__(self, bin_dir=None, output_invocation_dir=None):
        """
        :param bin_dir: path to the directory that contains the
          ldc_encoder, ldc_decoder and ldc_header_tool binaries. If it is None,
          options.external_bin_base_dir is None. If this is None as well, the
          same directory of this script is used by default.
        :output_invocation_dir: if not None, a path to a directory where invocations are to be stored.
        """
        bin_dir = bin_dir if bin_dir is not None else options.external_bin_base_dir
        bin_dir = bin_dir if bin_dir is not None else os.path.dirname(__file__)
        assert os.path.isdir(bin_dir), f"Invalid binary dir {bin_dir}."
        jpeg_bin_path = os.path.join(bin_dir, "jpeg")

        param_dict = sortedcontainers.SortedDict()
        icompression.WrapperCodec.__init__(
            self, compressor_path=jpeg_bin_path, decompressor_path=jpeg_bin_path,
            param_dict=param_dict, output_invocation_dir=output_invocation_dir)

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        s = " ".join(f"-{k} {v}" for k, v in self.param_dict.items())
        s += f" {original_path} {compressed_path}"
        return s

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"{compressed_path} {reconstructed_path}"


class JPEG(Abstract_JPEG, icompression.LossyCodec):
    """Compress unsigned 8-bit or 16-bit images using classic JPEG.
    Images must have 1 or 3 components
    """

    def __init__(self, quality=100, bin_dir=None, output_invocation_dir=None):
        """
        :param max_error: maximum pixelwise error allowed. Use 0 for lossless
          compression
        :param bin_dir: path to the directory that contains the
          ldc_encoder, ldc_decoder and ldc_header_tool binaries. If it is None,
          options.external_bin_base_dir is None. If this is None as well, the
          same directory of this script is used by default.
        :output_invocation_dir: if not None, a path to a directory where invocations are to be stored.
        """
        assert int(quality) == quality, f"quality must be an integer ({quality=})"
        assert 1 <= quality <= 100, f"quality must be an integer between 1 and 100 ({quality=})"
        super().__init__(bin_dir=bin_dir, output_invocation_dir=output_invocation_dir)
        self.param_dict["q"] = quality

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        assert original_file_info["bytes_per_sample"] in [1, 2], \
            f"Only 1 or 2 bytes per sample, unsigned values {original_file_info['bytes_per_sample']}"
        assert original_file_info["width"] <= self.max_dimension_size, \
            f"The input path has width {original_file_info['width']} exceeding the maximum {self.max_dimension_size}"
        assert original_file_info["height"] <= self.max_dimension_size, \
            f"The input path has height {original_file_info['height']} exceeding the maximum {self.max_dimension_size}"
        complete_array = isets.load_array_bsq(file_or_path=original_path, image_properties_row=original_file_info)
        assert not original_file_info["signed"], f"Only unsigned data can be compressed using JPEG"

        with tempfile.TemporaryDirectory() as tmp_dir:
            if complete_array.shape[2] == 1:
                # Grayscale image
                pbm_path = os.path.join(tmp_dir, "out.pgm")
                enb.pgm.write_pgm(array=complete_array[:, :, 0],
                                  bytes_per_sample=original_file_info["bytes_per_sample"],
                                  output_path=pbm_path)
            elif complete_array.shape[2] == 3:
                # RGB array
                pbm_path = os.path.join(tmp_dir, "out.ppm")
                enb.pgm.write_ppm(array=complete_array,
                                  bytes_per_sample=original_file_info["bytes_per_sample"],
                                  output_path=pbm_path)
            else:
                raise ValueError("Only 1-component or 3-component images can be compressed using classic JPEG. "
                                 f"{complete_array.shape=}")

            compression_results = super().compress(original_path=pbm_path, compressed_path=compressed_path)
            compression_results.original_path = original_path
            return compression_results

    def decompress(self, compressed_path: str, reconstructed_path: str, original_file_info=None):
        with tempfile.TemporaryDirectory() as tmp_dir:
            if original_file_info["component_count"] == 1:
                pbm_path = os.path.join(tmp_dir, "out.pgm")
            elif original_file_info["component_count"] == 3:
                pbm_path = os.path.join(tmp_dir, "out.ppm")
            else:
                raise ValueError("Only 1-component or 3-component images can be decompressed using classic JPEG")

            decompression_results = super().decompress(
                compressed_path=compressed_path, reconstructed_path=pbm_path)
            if original_file_info["component_count"] == 1:
                array = enb.pgm.read_pgm(pbm_path)
                array = array.reshape((array.shape[0], array.shape[1], 1))
            elif original_file_info["component_count"] == 3:
                array = enb.pgm.read_ppm(pbm_path)
            else:
                raise ValueError("Only 1-component or 3-component images can be decompressed using classic JPEG")
            assert len(array.shape) == 3, f"Invalid array shape ({array.shape=})"
            assert array.shape[2] in (1, 3), f"Invalid array 3rd-D length ({array.shape=})"

            enb.isets.dump_array_bsq(array=array, file_or_path=reconstructed_path)
            decompression_results.reconstructed_path = reconstructed_path
            return decompression_results

    @property
    def label(self):
        return f"JPEG Q={self.param_dict['q']}"


class JPEG_LS(Abstract_JPEG, icompression.LosslessCodec, icompression.NearLosslessCodec):
    """Compress unsigned 8-bit or 16-bit images using JPEG-LS.
    If images contain more than one component, they are combined into a single image 
    by vertically appending the different components.
    If the maximum dimension size allowed by JPEG-LS is exceeded, several images 
    are are produced in this way by splitting the input bands into the minimum 
    number of groups so that this limit is not exceeded.
    """

    def __init__(self, max_error=0, bin_dir=None, output_invocation_dir=None):
        """
        :param max_error: maximum pixelwise error allowed. Use 0 for lossless
          compression
        :param bin_dir: path to the directory that contains the
          ldc_encoder, ldc_decoder and ldc_header_tool binaries. If it is None,
          options.external_bin_base_dir is None. If this is None as well, the
          same directory of this script is used by default.
        :output_invocation_dir: if not None, a path to a directory where invocations are to be stored.
        """
        super().__init__(bin_dir=bin_dir, output_invocation_dir=output_invocation_dir)
        self.param_dict["ls"] = 0
        self.param_dict["c"] = ""
        max_error = int(max_error)
        assert max_error >= 0
        self.param_dict["m"] = int(max_error)

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        assert original_file_info["bytes_per_sample"] in [1, 2], \
            f"Only 1 or 2 bytes per sample, unsigned values {original_file_info['bytes_per_sample']}"
        assert original_file_info["width"] <= self.max_dimension_size, \
            f"The input path has width {original_file_info['width']} exceeding the maximum {self.max_dimension_size}"
        assert original_file_info["height"] <= self.max_dimension_size, \
            f"The input path has height {original_file_info['height']} exceeding the maximum {self.max_dimension_size}"
        complete_array = isets.load_array_bsq(file_or_path=original_path, image_properties_row=original_file_info)
        if original_file_info["signed"]:
            complete_array = complete_array.astype(np.int64)
            complete_array += 2 ** ((8 * original_file_info['bytes_per_sample']) - 1)
            assert (complete_array >= 0).all(), f"Error converting signed into unsigned"

        bands_per_image = min(original_file_info["component_count"],
                              self.max_dimension_size // original_file_info["height"])

        tw = tarlite.TarliteWriter()

        total_compression_time = 0
        maximum_memory_kb = -1

        with tempfile.TemporaryDirectory(dir=options.base_tmp_dir) as tmp_dir:
            stacked_size = 0
            for stack_index, start_band_index in enumerate(
                    range(0, original_file_info["component_count"], bands_per_image)):
                end_band_index_not_included = min(original_file_info["component_count"],
                                                  start_band_index + bands_per_image)
                stacked_array = np.hstack(tuple(complete_array[:, :, z].squeeze()
                                                for z in range(start_band_index, end_band_index_not_included)))
                stacked_size += stacked_array.shape[0] * stacked_array.shape[1]

                with tempfile.NamedTemporaryFile(dir=options.base_tmp_dir,
                                                 suffix=".pgm") as tmp_stack_file:
                    pgm.write_pgm(array=stacked_array,
                                  bytes_per_sample=original_file_info["bytes_per_sample"],
                                  output_path=tmp_stack_file.name)
                    stack_compressed_path = os.path.join(tmp_dir, str(stack_index))
                    compression_results = super().compress(
                        original_path=tmp_stack_file.name,
                        compressed_path=stack_compressed_path,
                        original_file_info=original_file_info)
                    total_compression_time += compression_results.compression_time_seconds
                    maximum_memory_kb = max(maximum_memory_kb, compression_results.maximum_memory_kb)

                    tw.add_file(input_path=stack_compressed_path)
            assert stacked_size == complete_array.shape[0] * complete_array.shape[1] * complete_array.shape[2], \
                f"Total stacked size {stacked_size} does not match the expectations. " \
                f"({stacked_size} vs {complete_array.shape[0] * complete_array.shape[1] * complete_array.shape[2]}"

            tw.write(output_path=compressed_path)

            compression_results = self.compression_results_from_paths(
                original_path=original_path, compressed_path=compressed_path)
            compression_results.compression_time_seconds = total_compression_time
            compression_results.maximum_memory_kb = maximum_memory_kb

            return compression_results

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        tr = tarlite.TarliteReader(tarlite_path=compressed_path)

        bands_per_image = min(original_file_info["component_count"],
                              self.max_dimension_size // original_file_info["height"])

        total_decompression_time = 0
        maximum_memory_kb = -1
        recovered_components = []
        with tempfile.TemporaryDirectory(dir=options.base_tmp_dir) as tmp_dir:
            tr.extract_all(output_dir_path=tmp_dir)

            for stack_index, start_band_index in enumerate(
                    range(0, original_file_info["component_count"], bands_per_image)):
                compressed_stack_path = os.path.join(tmp_dir, str(stack_index))

                with tempfile.NamedTemporaryFile(dir=options.base_tmp_dir, suffix=".pgm") as stack_path:
                    decompression_results = super().decompress(
                        compressed_path=compressed_stack_path, reconstructed_path=stack_path.name)
                    total_decompression_time += decompression_results.decompression_time_seconds
                    maximum_memory_kb = max(maximum_memory_kb, decompression_results.maximum_memory_kb)

                    assert os.path.isfile(stack_path.name)
                    stack_array = pgm.read_pgm(input_path=stack_path.name)

                    limits = tuple(range(original_file_info["height"],
                                         stack_array.shape[1],
                                         original_file_info["height"]))

                    recovered_components.extend(np.hsplit(stack_array, limits))

        e_sum = 0
        for c in recovered_components:
            assert len(c.shape) == 2
            e_sum += c.shape[0] * c.shape[1]
        assert e_sum == original_file_info['width'] * original_file_info['height'] * original_file_info[
            'component_count'], \
            f"Wrong number of recovered pixels {e_sum}, " \
            f"expected {original_file_info['width'] * original_file_info['height'] * original_file_info['component_count']}."
        assert len(recovered_components) == original_file_info["component_count"]
        assert recovered_components[0].shape == (original_file_info["width"], original_file_info["height"])

        recovered_array = np.dstack(recovered_components)
        assert recovered_array.shape == (
            original_file_info['width'], original_file_info['height'], original_file_info['component_count'])

        if original_file_info["signed"]:
            recovered_array = recovered_array.astype(np.int64)
            recovered_array -= 2 ** ((8 * original_file_info["bytes_per_sample"]) - 1)

        isets.dump_array_bsq(array=recovered_array, file_or_path=reconstructed_path,
                             dtype=isets.iproperties_row_to_numpy_dtype(image_properties_row=original_file_info))

        decompression_results = self.decompression_results_from_paths(
            compressed_path=compressed_path, reconstructed_path=reconstructed_path)
        decompression_results.decompression_time_seconds = total_decompression_time
        decompression_results.maximum_memory_kb = maximum_memory_kb
        return decompression_results

    @property
    def label(self):
        return f"JPEG-LS{' PAE ' + str(self.param_dict['m']) if self.param_dict['m'] > 0 else ''}"



import enb.plugins


class LCFrameworkPlugin(enb.plugins.PluginMake):
    name = "lc"
    label = "LC Framework codec generation and application tools"
    tags = {"data compression", "codec"}
    authors = ["Pau Quintas Torra", "Xavier Fernandez Mellado"]
    contrib_authors = ["Martin Burtscher"]
    contrib_reference_urls = ["https://github.com/burtscher/LC-framework"]
    contrib_download_url_name = [
        ("https://github.com/xavifeme00/LC-framework/archive/refs/heads/main.zip?raw=true",
         "lc-framework.zip")]
    tested_on = {"linux"}
